# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# *************************************  EMStorageMechanism **********************************************

"""

Contents
--------

  * `EMStorageMechanism_Overview`
    - `EMStorageMechanism_Memory`
    - `EMStorageMechanism_Entry`
    - `EMStorageMechanism_Fields`
  * `EMStorageMechanism_Creation`
  * `EMStorageMechanism_Structure`
  * `EMStorageMechanism_Execution`
  * `EMStorageMechanism_Class_Reference`


.. _EMStorageMechanism_Overview:

Overview
--------

An EMStorageMechanism is a subclass of `LearningMechanism`, modified for use in an `EMComposition` to store a new
entry in its `memory <EMComposition.memory>` attribute each time it executes.

.. _EMStorageMechanism_Memory:

# FIX: NEEDS EDITING:

* **Memory** -- the `memory <EMComposition.memory>` attribute of an `EMComposition` is a list of entries, each of
    which is a 2d np.array with a shape that corresponds to the `memory_matrix <EMStorageMechanism.memory_matrix>`
    attribute of the EMStorageMechanism that stores it.  Each entry is stored in the `memory <EMComposition.memory>`
    attribute of the EMComposition as a row or column of the `matrix <MappingProjection.matrix>` parameter of the
    `MappingProjections <MappingProjection>` to which the `LearningProjections <LearningProjection>` of the
    EMStorageMechanism project.  The `memory <EMComposition.memory>` attribute of the EMComposition is used by its
    `controller <EMComposition.controller>` to generate the `memory <EMMemoryMechanism.memory>` attribute of an
    `EMMemoryMechanism` that is used to retrieve entries from the `memory <EMComposition.memory>` attribute of the
    EMComposition.

.. _EMStorageMechanism_Entry:

* **Entry** -- an entry is a 2d np.array with a shape that corresponds to the `memory_matrix
    <EMStorageMechanism.memory_matrix>` attribute of the EMStorageMechanism that stores it.  Each entry is stored in the
    `memory <EMComposition.memory>` attribute of the EMComposition as a row or column of the `matrix
    <MappingProjection.matrix>` parameter of the `MappingProjections <MappingProjection>` to which the
    `LearningProjections <LearningProjection>` of the EMStorageMechanism project.  The `memory
    <EMComposition.memory>` attribute of the EMComposition is used by its `controller <EMComposition.controller>` to
    generate the `memory <EMMemoryMechanism.memory>` attribute of an `EMMemoryMechanism` that is used to retrieve
    entries from the `memory <EMComposition.memory>` attribute of the EMComposition.

.. _EMStorageMechanism_Fields:

* **Fields** -- an entry is composed of one or more fields, each of which is a 1d np.array with a length that
    corresponds to the number of `fields <EMStorageMechanism_Fields>` of the EMStorageMechanism that stores it.  Each
    field is stored in the `memory <EMComposition.memory>` attribute of the EMComposition as a row or column of the
    `matrix <MappingProjection.matrix>` parameter of the `MappingProjections <MappingProjection>` to which the
    `LearningProjections <LearningProjection>` of the EMStorageMechanism project.  The `memory
    <EMComposition.memory>` attribute of the EMComposition is used by its `controller <EMComposition.controller>` to
    generate the `memory <EMMemoryMechanism.memory>` attribute of an `EMMemoryMechanism` that is used to retrieve
    entries from the `memory <EMComposition.memory>` attribute of the EMComposition.

.. _EMStorageMechanism_Creation:

Creating an EMStorageMechanism
--------------------------------------------

An EMStorageMechanism can be created directly by calling its constructor, but most commonly it is created
automatically when an `EMComposition` is created, as its `learning_mechanism <EMComposition.learning_mechanism>`
used to store entries in its `memory <EMComposition.memory>` of the EMComposition. The `memory_matrix` must be
specified (as a template for the shape of the entries to be stored, and of the `matrix <MappingProjection.matrix>`
parameters to which they are assigned. It must also have at least one, and usually several `fields
<EMStorageMechanism.fields>` specifications that identify the `OutputPort`\\s of the `ProcessingMechanism`\\s from
which it receives its `fields <EMStorageMechanism_Fields>`, and a `field_types <EMStorageMechanism.field_types>`
specification that indicates whether each `field is a key or a value field <EMStorageMechanism_Fields>`.

.. _EMStorageMechanism_Structure:

Structure
---------

An EMStorageMechanism differs from a standard `LearningMechanism` in the following ways:

  * it has no `input_source <LearningMechanism.input_source>`, `output_source <LearningMechanism.output_source>`,
    or `error_source <LearningMechanism.error_source>` attributes;  instead, it has the `fields
    <EMStorageMechanism.fields>` and `field_types <EMStorageMechanism.field_types>` attributes described below.

  * its `fields <EMStorageMechanism.fields>` attribute has as many *FIELDS* `field <EMStorage_mechanism.fields>`
    as there are `fields <EMStorageMechanism_Fields>` of an entry in its `memory_matrix
    <EMStorageMechanism.memory_matrix>` attribute;  these are listed in its `fields <EMStorageMechanism.fields>`
    attribute and serve as the `InputPort`\\s for the EMStorageMechanism;  each receives a `MappingProjection` from
    the `OutputPort` of a `ProcessingMechanism`, the activity of which constitutes the corresponding `field
    <EMStorageMechanism_Fields>` of the `entry <EMStorageMechanism_Entry>` to be stored in its `memory_matrix
    <EMStorageMechanism.memory_matrix>` attribute.

  * it has a `field_types <EMStorageMechanism.field_types>` attribute that specifies whether each `field
    <EMStorageMechanism_Fields>` is a `key or a value field <EMStorageMechanism_Fields>`.

  * it has a `field_weights <EMStorageMechanism.field_weights>` attribute that specifies whether each `field
    <EMStorageMechanism_Fields>` each norms for each field are weighted before deteterming the weakest `entry
    <EMStorageMechanism_Entry>` in `memory_matrix <EMStorageMechanism.memory_matrix>`.

  * it has a `memory_matrix <EMStorageMechanism.memory_matrix>` attribute that represents the full memory that the
    EMStorageMechanism is used to update.

  * it has a `concatenation_node <EMStorageMechanism.concatenation_node>` attribute used to access the concatenated
    inputs to the `key <EMStorageMechanism.key>` fields of the `entry <EMStorageMechanism_Entry>` to be stored in its
    `memory_matrix <EMStorageMechanism.memory_matrix>` attribute.

  * it has a several *LEARNING_SIGNAL* `OutputPorts <OutputPort>` that each send a `LearningProjection` to the `matrix
    <MappingProjection.matrix>` parameter of a 'MappingProjection` that constitutes a `field <EMStorageMechanism_Fields>`
    of the `memory_matrix <EMStorageMechanism.memory_matrix>` attribute.

  * its `function <EMStorageMechanism.function>` is an `EMStorage` `LearningFunction`, that takes as its `variable
    <Function_Base.variable>` a list or 1d np.array with a length of the corresponding  *ACTIVATION_INPUT* InputPort;
    and it returns a `learning_signal <LearningMechanism.learning_signal>` (a weight matrix assigned to one of the
    Mechanism's *LEARNING_SIGNAL* OutputPorts), but no `error_signal <LearningMechanism.error_signal>`.

  * the default form of `modulation <ModulatorySignal_Modulation>` for its `learning_signals
    <LearningMechanism.learning_signals>` is *OVERRIDE*, so that the `matrix <MappingProjection.matrix>` parameter of
    the `MappingProjection` to which the `LearningProjection` projects is replaced by the `value
    <LearningProjection.value>` of the `learning_signal <LearningMechanism.learning_signal>`.

  * its `decay_rate <EMStorageMechanism.decay_rate>`, a float in the interval [0,1] that is used to decay
    `memory_matrix <EMStorageMechanism.memory_matrix>` before an `entry <EMStorageMechanism_Entry>` is stored.

  * its `storage_prob <EMStorageMechanism.storage_prob>`, a float in the interval [0,1] is used in place of a
    LearningMechanism's `storage_prob <LearningMechanism.storage_prob>` to determine the probability that the
    Mechanism will store its `variable <EMStorageMechanism.variable>` in its `memory_matrix
    <EMStorageMechanism.memory_matrix>` attribute each time it executes.

.. _EMStorageMechanism_Execution:

Execution
---------

An EMStorageMechanism executes after all of the other Mechanisms in the `EMComposition` to which it belongs have
executed.  It executes in the same manner as standard `LearningMechanism`, however instead of modulating
the `matrix <MappingProjection.matrix>` Parameter of a `MappingProjection`, it replaces a row or column in each of
the `matrix <MappingProjection.matrix>` Parameters of the `MappingProjections <MappingProjection>` to which its
`LearningProjections <LearningProjection>` project with an item of its `variable <EMStorageMechanism.variable>` that
represents the corresponding `field <EMStorageMechanism.fields>`. The entry replaced is the one that has the lowest
norm computed across all `fields <EMSorageMechanism_Fields>` of the `entry <EMStorageMechanism_Entry>` weighted by the
corresponding items of `field_weights <EMStorageMechanism.field_weights>` if that is specified.


.. _EMStorageMechanism_Class_Reference:

Class Reference
---------------

"""

import numpy as np
import re
from beartype import beartype

from psyneulink._typing import Optional, Union, Callable, Literal

from psyneulink.core.components.component import parameter_keywords
from psyneulink.core.components.functions.nonstateful.learningfunctions import EMStorage
from psyneulink.core.components.mechanisms.mechanism import Mechanism
from psyneulink.core.components.mechanisms.modulatory.learning.learningmechanism import \
    LearningMechanism, LearningMechanismError, LearningTiming, LearningType
from psyneulink.core.components.projections.projection import Projection, projection_keywords
from psyneulink.core.components.ports.parameterport import ParameterPort
from psyneulink.core.components.ports.outputport import OutputPort
from psyneulink.core.globals.context import ContextFlags
from psyneulink.core.globals.keywords import \
    (ADDITIVE, EM_STORAGE_MECHANISM, LEARNING, LEARNING_PROJECTION, LEARNING_SIGNALS, MULTIPLICATIVE,
     MULTIPLICATIVE_PARAM, MODULATION, NAME, OVERRIDE, OWNER_VALUE, PROJECTIONS, REFERENCE_VALUE, VARIABLE)
from psyneulink.core.globals.parameters import Parameter, check_user_specified, FunctionParameter, copy_parameter_value
from psyneulink.core.globals.preferences.basepreferenceset import ValidPrefSet
from psyneulink.core.globals.preferences.preferenceset import PreferenceLevel
from psyneulink.core.globals.utilities import convert_all_elements_to_np_array, is_numeric, all_within_range

__all__ = [
    'EMStorageMechanism', 'EMStorageMechanismError',
]

# Parameters:

parameter_keywords.update({LEARNING_PROJECTION, LEARNING})
projection_keywords.update({LEARNING_PROJECTION, LEARNING})

MEMORY_MATRIX = 'memory_matrix'
FIELDS = 'fields'
FIELD_TYPES = 'field_types'

def _memory_matrix_getter(owning_component=None, context=None)->list:
    """Return list of memories in which rows (outer dimension) are memories for each field.
    These are derived from `matrix <MappingProjection.matrix>` parameter of the `afferent
    <Mechanism_Base.afferents>` MappingProjections to each of the `retrieved_nodes <EMComposition.retrieved_nodes>`.
    """
    if owning_component.is_initializing:
        if owning_component.learning_signals is None or owning_component.input_ports is None:
            return None

    num_fields = len(owning_component.input_ports)

    # Get learning_signals that project to retrieved_nodes
    num_learning_signals = len(owning_component.learning_signals)
    learning_signals_for_retrieved = owning_component.learning_signals[num_learning_signals - num_fields:]

    # Get memory from learning_signals that project to retrieved_nodes
    if owning_component.is_initializing:
        # If initializing, learning_signals are still MappingProjections used to specify them, so get from them
        memory = [retrieved_learning_signal.parameters.matrix._get(context)
                  for retrieved_learning_signal in learning_signals_for_retrieved]
    else:
        # Otherwise, get directly from the learning_signals
        memory = [retrieved_learning_signal.efferents[0].receiver.owner.parameters.matrix._get(context)
                  for retrieved_learning_signal in learning_signals_for_retrieved]

    # Get memory capacity from first length of first matrix (can use full set since might be ragged array)
    memory_capacity = len(memory[0])

    # Reorganize memory so that each row is an entry and each column is a field
    return convert_all_elements_to_np_array([
        [memory[j][i] for j in range(num_fields)]
        for i in range(memory_capacity)
    ])


class EMStorageMechanismError(LearningMechanismError):
    pass


class EMStorageMechanism(LearningMechanism):
    """
    EMStorageMechanism(                       \
        variable,                             \
        fields,                               \
        field_types,                          \
        memory_matrix,                        \
        function=EMStorage,                   \
        storage_prob=1.0,                     \
        decay_rate=0.0,                       \
        learning_signals,                     \
        modulation=OVERRIDE,                  \
        params=None,                          \
        name=None,                            \
        prefs=None)

    Implements a `LearningMechanism` that modifies the `matrix <MappingProjection.matrix>` parameters of
    `MappingProjections <MappingProjection>` that implement its `memory_matrix <EMStorageMechanism.memory_matrix>`.

    Arguments
    ---------

    variable : List or 2d np.array : default None
        each item of the 2d array specifies the shape of the corresponding `field <EMStorageMechanism_Fields>` of
        an `entry <EMStorageMechanism_Entry>`, that must be compatible (in number and type) with the `value
        <InputPort.value>` of the corresponding item of its `fields <EMStorageMechanism.fields>`
        attribute (see `variable <EMStorageMechanism.variable>` for additional details).

    fields : List[OutputPort, Mechanism, Projection, tuple[str, Mechanism, Projection] or dict] : default None
        specifies the `OutputPort`\\(s), the `value <OutputPort.value>`\\s of which are used as the
        corresponding `fields <EMStorageMechanism_Fields>` of the `memory_matrix <EMStorageMechanism.memory_matrix>`;
        used to construct the Mechanism's `InputPorts <InputPort>`; must be the same length as `variable
        <EMStorageMechanism.variable>`.

    field_types : List[int] : default None
        specifies whether each item of `variable <EMStorageMechanism.variable>` corresponds to a `key or value field
        <EMStorageMechanism_Fields>` (see `field_types <EMStorageMechanism.field_types>` for additional details);
        must contain only 1's (for keys) and 0's (for values), with the same number of these as there are items in
        the `variable <EMStorageMechanism.variable>` and `fields <EMStorageMechanism.fields>` arguments.

    field_weights : List[float] : default None
        specifies whether norms for each field are weighted before determining the weakest `entry
        <EMStorageMechanism_Entry>` in `memory_matrix <EMStorageMechanism.memory_matrix>`. If None (the default),
        the norm of each `entry <EMStorageMechanism_Entry>` is calculated across all fields at once; if specified,
        it must contain only floats from 0 to 1, and be the same length as the `fields <EMStorageMechanism.fields>`
        argument (see `field_weights <EMStorageMechanism.field_types>` for additional details).

    concatenation_node : OutputPort or Mechanism : default None
        specifies the `OutputPort` or `Mechanism` in which the `value <OutputPort.value>` of the `key fields
        <EMStorageMechanism_Fields>` are concatenated (see `concatenate keys <EMComposition_Concatenate_Queries>`
        for additional details).

    memory_matrix : List or 2d np.array : default None
        specifies the shape of the `memory <EMStorageMechanism_Memory>` used to store an `entry
        <EMStorageMechanism_Entry>` (see `memory_matrix <EMStorageMechanism.memory_matrix>` for additional details).

    function : LearningFunction or function : default EMStorage
        specifies the function used to assign each item of the `variable <EMStorageMechanism.variable>` to the
        corresponding `field <EMStorageMechanism_Fields>` of the `memory_matrix <EMStorageMechanism.memory_matrix>`.
        It must take as its `variable <EMStorage.variable>` argument a list or 1d array of numeric values
        (the "activity vector"); a ``memory_matrix`` argument that is a 2d array to which
        the `variable <EMStorageMechanism.variable>` is assigned; ``axis`` and ``storage_location`` arguments that
        determine where in ``memory_matrix`` the `variable <EMStorageMechanism.variable>` is stored; and optional
        ``storage_prob`` and ``decay_rate`` arguments that determine the probability with which storage occurs and
        the rate at which the `memory_matrix <EMStorageMechanism.memory_matrix>` decays, respectively.  The function
        must return a list, 2d np.array for the corresponding `field <EMStorageMechanism_Fields>` of the
        `memory_matrix <EMStorageMechanism.memory_matrix>` that is updated (see `EMStorage` for additional details).

    learning_signals : List[ParameterPort, Projection, tuple[str, Projection] or dict] : default None
        specifies the `ParameterPort`\\(s) for the `matrix <MappingProjection.matrix>` parameter of the
        `MappingProjection>`\\s that implement the `memory <EMStorageMechanism_Memory>` in which the `entry
        <EMStorageMechanism_Entry>` is stored; there must the same number of these as `fields
        <EMStorageMechanism.fields>`, and they must be specified in the sqme order.

    modulation : str : default OVERRIDE
        specifies form of `modulation <ModulatorySignal_Modulation>` that `learning_signals
        <EMStorageMechanism.learning_signals>` use to modify the `matrix <MappingProjection.matrix>` parameter of the
        `MappingProjections <MappingProjection>` that implement the `memory <EMStorageMechanism_Memory>` in which
        `entries <EMStorageMechanism_Entry>` is stored (see `modulation <EMStorageMechanism_Modulation>` for additional
        details).

    storage_prob : float : default None
        specifies the probability with which the current entry is stored in the EMSorageMechanism's `memory_matrix
        <EMStorageMechanism.memory_matrix>` (see `storage_prob <EMStorageMechanism.storage_prob>` for details).

    decay_rate : float : default 0.0
        specifies the rate at which `entries <EMStorageMechanism_Entry>` in the `memory_matrix
        <EMStorageMechanism.memory_matrix>` decay (see `decay_rate <EMStorageMechanism.decay_rate>` for additional
        details).

    Attributes
    ----------

    # FIX: FINISH EDITING:

    variable : 2d np.array

        each item of the 2d array is used as a template for the shape of each the `fields
        <EMStorageMechanism_Fields>` that  comprise and `entry <EMStorageMechanism_Entry>` in the `memory_matrix
        <EMStorageMechanism.memory_matrix>`, and that must be compatible (in number and type) with the `value
        <OutputPort.value>` of the item specified the corresponding itme of its `fields <EMStorageMechanism.fields>`
        attribute. The values of the `variable <EMStorageMechanism.variable>` are assigned to the `memory_matrix
        <EMStorageMechanism.memory_matrix>` by the `function <EMStorageMechanism.function>`.

    fields : List[OutputPort, Mechanism, Projection, tuple[str, Mechanism, Projection] or dict] : default None
        the `OutputPort`\\(s) used to get the value for each `field <EMStorageMechanism_Fields>` of
        an `entry <EMStorageMechanism_Entry>` of the `memory_matrix <EMStorageMechanism.memory_matrix>` attribute.

    field_types : List[int or tuple[slice]]
        contains a list of indicators of whether each item of `variable <EMStorageMechanism.variable>`
        and the corresponding `fields <EMStorageMechanism.fields>` are key (1) or value (0) fields.
        (see `fields <EMStorageMechanism_Fields>` for additional details).

    field_weights : List[float] or None
        determines whether norms for each field are weighted before identifying the weakest `entry
        <EMStorageMechanism_Entry>` in `memory_matrix <EMStorageMechanism.memory_matrix>`. If is None (the default),
        the norm of each `entry <EMStorageMechanism_Entry>` is calculated across all fields at once; if specified,
        it must contain only floats from 0 to 1, and be the same length as the `fields <EMStorageMechanism.fields>`
        argument (see `field_weights <EMStorageMechanism.field_types>` for additional details).

    learned_projections : List[MappingProjection]
        list of the `MappingProjections <MappingProjection>`, the `matrix <MappingProjection.matrix>` Parameters of
        which are modified by the EMStorageMechanism.

    function : LearningFunction or function : default EMStorage
        the function used to assign the value of each `field <EMStorageMechanism.fields>` to the corresponding entry
        in `memory_matrix <EMStorageMechanism.memory_matrix>`.  It must take as its `variable <EMSorage.variable>`
        argument a list or 1d array of numeric values (an `entry
        <EMStorage.entry`) and return a list, 2d np.array assigned to
        the corresponding `field <EMStorageMechanism_Fields>` of the
        `memory_matrix <EMStorageMechanism.memory_matrix>`.

    storage_prob : float
        specifies the probability with which the current entry is stored in the EMSorageMechanism's `memory_matrix
        <EMStorageMechanism.memory_matrix>`.

    decay_rate : float : default 0.0
        determines the rate at which `entries <EMStorageMechanism_Entry>` in the `memory_matrix
        <EMStorageMechanism.memory_matrix>` decay;  the decay rate is applied to `memory_matrix
        <EMStorageMechanism.memory_matrix>` before it is updated with the new `entry <EMStorageMechanism_Entry>`.

    learning_signals : List[LearningSignal]
        list of all of the `LearningSignals <LearningSignal>` for the EMStorageMechanism, each of which
        sends a `LearningProjection` to the `ParameterPort`\\(s) for the `MappingProjections
        <MappingProjection>` that implement the `memory <EMStorageMechanism_Memory>` in which the `entry
        <EMStorageMechanism_Entry>` is stored.  The `value <LearningSignal.value>` of each LearningSignal is
        used by its `LearningProjection` to modify the `matrix <MappingProjection.matrix>` parameter of the
        MappingProjection to which that projects.

    learning_projections : List[LearningProjection]
        list of all of the LearningProjections <LearningProjection>` from the EMStorageMechanism, listed
        in the order of the `LearningSignals <LearningSignal>` to which they belong (that is, in the order they are
        listed in the `learning_signals <EMStorageMechanism.learning_signals>` attribute).

    modulation : str
        determines form of `modulation <ModulatorySignal_Modulation>` that `learning_signals
        <EMStorageMechanism.learning_signals>` use to modify the `matrix <MappingProjection.matrix>` parameter of the
        `MappingProjections <MappingProjection>` that implement the `memory <EMStorageMechanism_Memory>` in which
        `entries <EMStorageMechanism_Entry>` is stored.  *OVERRIDE* (the default) insure that entries are stored
        exactly as specified by the `value <OutputPort.value>` of the `fields <EMStorageMechanism.fields>` of the
        `entry <EMStorageMechanism_Entry>`;  other values can have unpredictable consequences
        (see `ModulatorySignal_Types for additional details)

    output_ports : ContentAddressableList[OutputPort]
        list of the EMStorageMechanism's `OutputPorts <OutputPort>`, beginning with its
        `learning_signals <EMStorageMechanism.learning_signals>`, and followed by any additional
        (user-specified) `OutputPorts <OutputPort>`.

    output_values : 2d np.array
        the first items are the `value <OutputPort.value>`\\(s) of the LearningMechanism's `learning_signal
        <EMStorageMechanism.learning_signal>`\\(s), followed by the `value <OutputPort.value>`(s)
        of any additional (user-specified) OutputPorts.

    """

    componentType = EM_STORAGE_MECHANISM
    className = componentType
    suffix = " " + className

    class Parameters(LearningMechanism.Parameters):
        """
            Attributes
            ----------

                concatenation_node
                    see `concatenation_node <EMStorageMechanism.concatenation_node>`

                    :default value: None
                    :type: ``Mechanism or OutputPort``
                    :read only: True

                decay_rate
                    see `decay_rate <EMStorageMechanism.decay_rate>`

                    :default value: 0.0
                    :type: ``float``

                fields
                    see `fields <EMStorageMechanism.fields>`

                    :default value: None
                    :type: ``list``
                    :read only: True

                field_types
                    see `field_types <EMStorageMechanism.field_types>`

                    :default value: None
                    :type: ``list``
                    :read only: True

                field_weights
                    see `field_weights <EMStorageMechanism.field_weights>`

                    :default value: None
                    :type: ``list or np.ndarray``

                memory_matrix
                    see `memory_matrix <EMStorageMechanism.memory_matrix>`

                    :default value: None
                    :type: ``np.ndarray``
                    :read only: True

                function
                    see `function <EMStorageMechanism.function>`

                    :default value: `EMStorage`
                    :type: `Function`

                input_ports
                    see `fields <EMStorageMechanism.fields>`

                    :default value: None
                    :type: ``list``
                    :read only: True

                learning_signals
                    see `learning_signals <EMStorageMechanism.learning_signals>`

                    :default value: []
                    :type: ``List[MappingProjection or ParameterPort]``
                    :read only: True

                modulation
                    see `modulation <EMStorageMechanism.modulation>`

                    :default value: OVERRIDE
                    :type: ModulationParam
                    :read only: True

                output_ports
                    see `learning_signals <EMStorageMechanism.learning_signals>`

                    :default value: None
                    :type: ``list``
                    :read only: True

                storage_prob
                    see `storage_prob <EMStorageMechanism.storage_prob>`

                    :default value: 1.0
                    :type: ``float``

        """
        input_ports = Parameter([], # FIX: SHOULD BE ABLE TO UE THIS WITH 'fields' AS CONSTRUCTOR ARGUMENT
                                stateful=False,
                                loggable=False,
                                read_only=True,
                                structural=True,
                                parse_spec=True,
                                constructor_argument='fields',
                                )
        fields = Parameter(
            [], stateful=False, loggable=False, read_only=True, structural=True
        )
        field_types = Parameter([],stateful=False,
                                loggable=False,
                                read_only=True,
                                structural=True,
                                parse_spec=True,
                                dependiencies='fields')
        field_weights = Parameter(None,
                                  modulable=True,
                                  stateful=True,
                                  loggable=True,
                                  dependiencies='fields')
        concatenation_node = Parameter(None,
                                       stateful=False,
                                       loggable=False,
                                       read_only=True,
                                       structural=True)
        function = Parameter(EMStorage, stateful=False, loggable=False)
        # storage_prob = Parameter(1.0, modulable=True, stateful=True)
        storage_prob = FunctionParameter(1.0,
                                         function_name='function',
                                         function_parameter_name='storage_prob',
                                         primary=True,
                                         modulable=True,
                                         aliases=[MULTIPLICATIVE_PARAM],
                                         stateful=True)
        decay_rate = Parameter(0.0, modulable=True, stateful=True)
        memory_matrix = Parameter(None, getter=_memory_matrix_getter, read_only=True, structural=True)
        modulation = OVERRIDE
        output_ports = Parameter([],
                                 stateful=False,
                                 loggable=False,
                                 read_only=True,
                                 structural=True,
                                 # constructor_argument='learning_signals'
                                 )
        learning_signals = Parameter([],
                                     stateful=False,
                                     loggable=False,
                                     read_only=True,
                                     structural=True)
        learning_type = LearningType.UNSUPERVISED
        # learning_type = LearningType.SUPERVISED
        # learning_timing = LearningTiming.LEARNING_PHASE
        learning_timing = LearningTiming.EXECUTION_PHASE

    def _validate_field_types(self, field_types):
        if not len(field_types) or len(field_types) != len(self.input_ports):
            return f"must be specified with a number of items equal to " \
                   f"the number of fields specified {len(self.input_ports)}"
        if not all(item in {1,0} for item in field_types):
            return f"must be a list of 1s (for keys) and 0s (for values)."

    def _validate_field_weights(self, field_weights):
        if not field_weights or len(field_weights) != len(self.input_ports):
            return f"must be specified with a number of items equal to " \
                   f"the number of fields specified {len(self.input_ports)}"
        if not all(isinstance(item, (int, float)) and (0 <= item  <= 1) for item in field_weights):
            return f"must be a list floats from 0 to 1."

    def _validate_storage_prob(self, storage_prob):
        storage_prob = float(storage_prob)
        if not all_within_range(storage_prob, 0, 1):
            return f"must be a float in the interval [0,1]."

    def _validate_decay_rate(self, decay_rate):
        decay_rate = float(decay_rate)
        if not all_within_range(decay_rate, 0, 1):
            return f"must be a float in the interval [0,1]."


    classPreferenceLevel = PreferenceLevel.TYPE

    @check_user_specified
    @beartype
    def __init__(self,
                 default_variable: Union[list, np.ndarray],
                 fields: Union[list, tuple, dict, OutputPort, Mechanism, Projection] = None,
                 field_types: list = None,
                 field_weights: Optional[Union[list, np.ndarray]] = None,
                 concatenation_node: Optional[Union[OutputPort, Mechanism]] = None,
                 memory_matrix: Union[list, np.ndarray] = None,
                 function: Optional[Callable] = EMStorage,
                 learning_signals: Union[list, dict, ParameterPort, Projection, tuple] = None,
                 modulation: Optional[Literal[OVERRIDE, ADDITIVE, MULTIPLICATIVE]] = OVERRIDE,
                 decay_rate: Optional[Union[int, float, np.ndarray]] = 0.0,
                 storage_prob: Optional[Union[int, float, np.ndarray]] = 1.0,
                 params=None,
                 name=None,
                 prefs: Optional[ValidPrefSet] = None,
                 **kwargs
                 ):

        super().__init__(default_variable=default_variable,
                         fields=fields,
                         field_types=field_types,
                         concatenation_node=concatenation_node,
                         memory_matrix=memory_matrix,
                         function=function,
                         learning_signals=learning_signals,
                         modulation=modulation,
                         decay_rate=decay_rate,
                         storage_prob=storage_prob,
                         field_weights=field_weights,
                         params=params,
                         name=name,
                         prefs=prefs,
                         **kwargs)

    def _validate_variable(self, variable, context=None):
        """Validate that variable has only one item: activation_input.
        """

        # Skip LearningMechanism._validate_variable in call to super(), as it requires variable to have 3 items
        variable = super(LearningMechanism, self)._validate_variable(variable, context)

        # Items in variable should be 1d and have numeric values
        if not (all(np.array(variable)[i].ndim == 1 for i in range(len(variable))) and is_numeric(variable)):
            raise EMStorageMechanismError(f"Variable for {self.name} ({variable}) must be "
                                          f"a list or 2d np.array containing 1d arrays with only numbers.")
        return variable

    def _validate_params(self, request_set, target_set=None, context=None):
        """Validate relationship of matrix, fields and field_types arguments"""

        # Ensure that the shape of variable is equivalent to an entry in memory_matrix
        if MEMORY_MATRIX in request_set:
            memory_matrix = request_set[MEMORY_MATRIX]
            # Items in variable should have the same shape as memory_matrix
            if memory_matrix[0].shape != np.array(self.variable).shape:
                raise EMStorageMechanismError(f"The 'variable' arg for {self.name} ({self.variable}) must be "
                                              f"a list or 2d np.array containing entries that have the same shape "
                                              f"({memory_matrix.shape}) as an entry (row) in 'memory_matrix' arg.")

        # Ensure the number of fields is equal to the number of items in variable
        if FIELDS in request_set:
            fields = request_set[FIELDS]
            if len(fields) != len(self.variable):
                raise EMStorageMechanismError(f"The 'fields' arg for {self.name} ({fields}) must have the same "
                                              f"number of items as its variable arg ({len(self.variable)}).")

        # Ensure the number of field_types is equal to the number of fields
        if FIELD_TYPES in request_set:
            field_types = request_set[FIELD_TYPES]
            if len(field_types) != len(fields):
                raise EMStorageMechanismError(f"The 'field_types' arg for {self.name} ({field_types}) must have "
                                              f"the same number of items as its 'fields' arg ({len(fields)}).")

        num_keys = len([i for i in field_types if i==1])
        concatenate_queries = 'concatenation_node' in request_set and request_set['concatenation_node'] is not None

        # Ensure the number of learning_signals is equal to the number of fields + number of keys
        if LEARNING_SIGNALS in request_set:
            learning_signals = request_set[LEARNING_SIGNALS]
            if concatenate_queries:
                num_match_fields = 1
            else:
                num_match_fields = num_keys
            if len(learning_signals) != num_match_fields + len(fields):
                raise EMStorageMechanismError(f"The number of 'learning_signals' ({len(learning_signals)}) specified "
                                              f"for  {self.name} must be the same as the number of items "
                                              f"in its variable ({len(self.variable)}).")

        # Ensure shape of learning_signals matches shapes of matrices for match nodes (i.e., either keys or concatenate)
        key_indices = [i for i, field_type in enumerate(field_types) if field_type == 1]
        for i, learning_signal in enumerate(learning_signals[:num_match_fields]):
            learning_signal_shape = learning_signal.parameters.matrix._get(context).shape
            if concatenate_queries:
                memory_matrix_field_shape = np.array([np.concatenate(row, dtype=object).flatten()
                                                      for row in memory_matrix[:,0:num_keys]]).T.shape
            else:
                memory_matrix_field_shape = np.array(memory_matrix[:,key_indices[i]].tolist()).T.shape
            assert learning_signal_shape == memory_matrix_field_shape, \
                f"The shape ({learning_signal_shape}) of the matrix for the Projection {learning_signal.name} " \
                f"used to specify learning signal {i} of {self.name} does not match the shape " \
                f"of the corresponding field {i} of its 'memory_matrix' {memory_matrix_field_shape})."
        # Ensure shape of learning_signals matches shapes of matrices for retrieval nodes (i.e., all input fields)
        for i, learning_signal in enumerate(learning_signals[num_match_fields:]):
            learning_signal_shape = learning_signal.parameters.matrix._get(context).shape
            memory_matrix_field_shape = np.array(memory_matrix[:,i].tolist()).shape
            assert learning_signal_shape == memory_matrix_field_shape, \
                f"The shape ({learning_signal.shape}) of the matrix for the Projection {learning_signal.name} " \
                f"used to specify learning signal {i} of {self.name} does not match the shape " \
                f"of the corresponding field {i} of its 'memory_matrix' {memory_matrix.shape})."

    def _instantiate_input_ports(self, input_ports=None, reference_value=None, context=None):
        """Override LearningMechanism to instantiate an InputPort for each field"""
        input_ports = [{NAME: f"QUERY_INPUT_{i}" if self.field_types[i] == 1 else f"VALUE_INPUT_{i}",
                        VARIABLE: self.variable[i],
                        PROJECTIONS: field}
                       for i, field in enumerate(self.input_ports)]
        return super()._instantiate_input_ports(input_ports=input_ports, context=context)

    def _instantiate_output_ports(self, output_ports=None, reference_value=None, context=None):
        learning_signal_dicts = []
        for i, learning_signal in enumerate(self.learning_signals):
            learning_signal_dicts.append({NAME: f"STORE TO {learning_signal.receiver.owner.name} MATRIX",
                                          VARIABLE: (OWNER_VALUE, i),
                                          REFERENCE_VALUE: self.value[i],
                                          MODULATION: self.modulation,
                                          PROJECTIONS: learning_signal.parameter_ports['matrix']})
        self.parameters.learning_signals._set(learning_signal_dicts, context)

        learning_signals = super()._instantiate_output_ports(context=context)


    def _parse_function_variable(self, variable, context=None):
        # Function expects a single field (one item of Mechanism's variable) at a time
        if self.initialization_status == ContextFlags.INITIALIZING:
            # During initialization, Mechanism's variable is its default_variable,
            # which has all field's worth of input, so need get a single one here
            return variable[0]
        # During execution, _execute passes only a entry (item of variable) at a time,
        #    so can just pass that along here
        return variable

    def _execute(self,
                 variable=None,
                 context=None,
                 runtime_params=None):
        """Execute EMStorageMechanism.function and return learning_signals

        For each node in query_input_nodes and value_input_nodes,
        assign its value to afferent weights of corresponding retrieved_node.
        - memory = matrix of entries made up vectors for each field in each entry (row)
        - memory_full_vectors = matrix of entries made up of vectors concatentated across all fields (used for norm)
        - entry_to_store = query_input or value_input to store
        - field_memories = weights of Projections for each field

        DIVISION OF LABOR BETWEEN MECHANISM AND FUNCTION:
        EMStorageMechanism._execute:
         - compute norms to find weakest entry in memory
         - compute storage_prob to determine whether to store current entry in memory
         - call function for each LearningSignal to decay existing memory and assign input to weakest entry
        EMStorage function:
         - decay existing memories
         - assign input to weakest entry (given index passed from EMStorageMechanism)

        :return: List[2d np.array] self.learning_signal
        """

        # FIX: SET LEARNING MODE HERE FOR SHOW_GRAPH

        decay_rate = self.parameters.decay_rate._get(context)      # modulable, so use getter
        storage_prob = self.parameters.storage_prob._get(context)  # modulable, so use getter
        field_weights = self.parameters.field_weights._get(context)  # modulable, so use getter
        concatenation_node = self.concatenation_node
        num_match_fields = 1 if concatenation_node else len([i for i in self.field_types if i==1])

        memory = self.parameters.memory_matrix._get(context)
        if memory is None or self.is_initializing:
            if self.is_initializing:
                # Return existing matrices for field_memories  # FIX: THE FOLLOWING DOESN'T TEST FUNCTION:
                return convert_all_elements_to_np_array([
                    learning_signal.receiver.path_afferents[0].parameters.matrix._get(context)
                    for learning_signal in self.learning_signals
                ])
            # Raise exception if not initializing and memory is not specified
            else:
                owner_string = ""
                if self.owner:
                    owner_string = " of " + self.owner.name
                raise EMStorageMechanismError(f"Call to {self.__class__.__name__} function {owner_string} "
                                              f"must include '{MEMORY_MATRIX}' in params arg.")

        # Get least used slot (i.e., weakest memory = row of matrix with lowest weights) computed across all fields
        field_norms = np.empty((len(memory),len(memory[0])))
        for row in range(len(memory)):
            for col in range(len(memory[0])):
                field_norms[row][col] = np.linalg.norm(memory[row][col])
        if field_weights is not None:
            field_norms *= field_weights
        row_norms = np.sum(field_norms, axis=1)
        # IMPLEMENTATION NOTE:
        #  the following will give the lowest index in case of a tie;
        #  this means that if memory is initialized with all zeros,
        #  it will be occupied in row order
        idx_of_weakest_memory = np.argmin(row_norms)

        value = []
        for i, field_projection in enumerate([learning_signal.efferents[0].receiver.owner
                                            for learning_signal in self.learning_signals]):
            if i < num_match_fields:
                # For match matrices,
                #   get entry to store from variable of Projection matrix (memory_field)
                #   to match_node in which memory will be stored (this is to accomodate concatenation_node)
                axis = 0
                entry_to_store = field_projection.parameters.variable._get(context)
                if concatenation_node is None:
                    assert np.all(entry_to_store == variable[i]),\
                        f"PROGRAM ERROR: misalignment between inputs and fields for storing them"
            else:
                # For retrieval matrices,
                #    get entry to store from variable (which has inputs to all fields)
                axis = 1
                entry_to_store = variable[i - num_match_fields]
            # Get matrix containing memories for the field from the Projection
            field_memory_matrix = field_projection.parameters.matrix._get(context)

            # pass in field_projection matrix to EMStorage function
            res = super(LearningMechanism, self)._execute(
                variable=entry_to_store,
                memory_matrix=copy_parameter_value(field_memory_matrix),
                axis=axis,
                storage_location=idx_of_weakest_memory,
                storage_prob=storage_prob,
                decay_rate=decay_rate,
                context=context,
                runtime_params=runtime_params
            )
            value.append(res)
            # assign modified field_memory_matrix back
            field_projection.parameters.matrix._set(res, context)
        return convert_all_elements_to_np_array(value)



from . import autoassociativeprojection
from . import maskedmappingprojection

from .autoassociativeprojection import *
from .maskedmappingprojection import *

__all__ = list(autoassociativeprojection.__all__)
__all__.extend(maskedmappingprojection.__all__)



#
# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
#
# *******************************************  SELECTION FUNCTIONS *****************************************************
"""

* `OneHot`

COMMENT:
* TBI Threshold
* `KWTA`
COMMENT

Functions that selects a subset of elements to maintain or transform, while nulling the others.

"""

__all__ = ['SelectionFunction', 'OneHot', 'max_vs_avg', 'max_vs_next']

import warnings

import numpy as np
from beartype import beartype

from psyneulink._typing import Optional, Literal

from psyneulink.core import llvm as pnlvm
from psyneulink.core.components.component import DefaultsFlexibility
from psyneulink.core.components.functions.function import (
    DEFAULT_SEED, Function, Function_Base, FunctionError,
    _random_state_getter, _seed_setter,
)
from psyneulink.core.globals.keywords import \
    (ALL, ARG_MAX, ARG_MAX_ABS, ARG_MAX_ABS_INDICATOR, ARG_MAX_INDICATOR,
     ARG_MIN, ARG_MIN_ABS, ARG_MIN_ABS_INDICATOR, ARG_MIN_INDICATOR,
     DETERMINISTIC, FIRST, LAST,
     MAX, MAX_ABS_INDICATOR, MAX_ABS_VAL, MAX_INDICATOR, MAX_VAL,
     MIN, MIN_ABS_INDICATOR, MIN_ABS_VAL, MIN_INDICATOR, MIN_VAL,
     MODE, ONE_HOT_FUNCTION, PREFERENCE_SET_NAME, PROB, PROB_INDICATOR,
     RANDOM, SELECTION_FUNCTION_TYPE)

from psyneulink.core.globals.parameters import Parameter, check_user_specified
from psyneulink.core.globals.preferences.basepreferenceset import \
    REPORT_OUTPUT_PREF, PreferenceEntry, PreferenceLevel, ValidPrefSet

mode_options = [DETERMINISTIC, PROB, PROB_INDICATOR,
                ARG_MAX, ARG_MAX_ABS, ARG_MAX_INDICATOR, ARG_MAX_ABS_INDICATOR,
                ARG_MIN,  ARG_MIN_ABS, ARG_MIN_INDICATOR, ARG_MIN_ABS_INDICATOR,
                MAX_VAL, MAX_ABS_VAL, MAX_INDICATOR, MAX_ABS_INDICATOR,
                MIN_VAL, MIN_ABS_VAL, MIN_INDICATOR, MIN_ABS_INDICATOR]
tie_options = [ALL, FIRST, LAST, RANDOM]

# FIX: IMPLEMENT AS Functions
def max_vs_next(x):
    x_part = np.partition(x, -2)
    max_val = x_part[-1]
    next = x_part[-2]
    return max_val - next


def max_vs_avg(x):
    x_part = np.partition(x, -2)
    max_val = x_part[-1]
    others = x_part[:-1]
    return max_val - np.mean(others)


class SelectionFunction(Function_Base):
    """Functions that selects a particular value to maintain or transform, while nulling the others.
    """
    componentType = SELECTION_FUNCTION_TYPE


class OneHot(SelectionFunction):
    """
    OneHot(                \
         default_variable, \
         mode=MAX_VAL,     \
         direction=MAX,    \
         abs_val=FALSE     \
         indicator=FALSE,  \
         tie=ALL,          \
         params=None,      \
         owner=None,       \
         name=None,        \
         prefs=None        \
         )

    Return an array with one non-zero value.

    .. _OneHot:

    `function <Selection.function>` returns an array the same length as the first item in `variable <OneHot.variable>`,
    with all of its values zeroed except one (unless there is a tie, which is handled as specified by **tie**); the
    following options can be used in any combination:

        * **mode**: determines how the non-zero value(s) in the array is (are) selected

            * *STANDARD*: value (or 1) for the element(s) with the maximum or minimum value(s) in the array,
              as specified by the options below; all other elements are zeroed;  this is the default.

            * *PROB*: value of probabilistically chosen element based on probabilities passed in second item
              of variable; if there is a tie, a single element is chosen probabilistically.

            * *PROB_INDICATOR*: same as *PROB* but chosen item is assigned a value of 1;
              if there is a tie, a single element is chosen probabilistically.

         * **direction**: *MAX* (default) or *MIN*
           determines whether the maximum or minimum value(s) in the array are selected.

         * **abs_val**: *False* (default) or *True*
           determines whether the absolute values of the elements in the array are used to
           select the maximum or minimum value(s).

         * **indicator**:  *False* (default) or *True*
           determines whether the selected values(s) is (are) replace with a value of 1.

         * **tie**: *ALL* (default), *FIRST*, *LAST* or *RANDOM*
           determines how a tie is handled when there is more than one element with the maximum or minimum value;

           *ALL*: selects all elements in the tie;

           *FIRST*: selects the value of the element with the lowest index;

           *LAST*: selects the value of the element with the lowest index;

           *RANDOM*: randomly selects one of the tied elements;

        The following convenience keywords can be used to specify particular combinations of options for the **mode**
        argument together with the **tie** argument (these are included mainly for backward compatibility):

            * *ARG_MAX*: signed value of a single element with the maximum signed value,
              or the one with lowest index if there is a tie.

            * *ARG_MAX_ABS*: absolute value of a single element with the maximum absolute value,
              or the one with lowest index if there is a tie.

            * *ARG_MAX_INDICATOR*: 1 in place of single element with maximum signed value,
              or the one with lowest index if there is a tie.

            * *ARG_MAX_ABS_INDICATOR*: 1 in place of single element with maximum absolute value,
              or the one with lowest index if there is a tie.

            * *MAX_VAL*: signed value of the element with the maximum signed value;
              if there is a tie, which elements are returned is determined by `tie_index <OneHot.tie_index>`.

            * *MAX_ABS_VAL*: absolute value of the element with the maximum absolute value;
              if there is a tie, which elements are returned is determined by `tie_index <OneHot.tie_index>`.

            * *MAX_INDICATOR*: 1 in place of the element with the maximum signed value;
              if there is a tie, which elements are returned is determined by `tie_index <OneHot.tie_index>`.

            * *MAX_ABS_INDICATOR*: 1 in place of the element(s) with the maximum absolute value;
              if there is a tie, which elements are returned is determined by `tie_index <OneHot.tie_index>`.

            * *ARG_MIN*: signed value of a single element with the minium signed value,
              or the one with lowest index if there is a tie.

            * *ARG_MIN_ABS*: absolute value of a single element with the minium absolute value,
              or the one with lowest index if there is a tie.

            * *ARG_MIN_INDICATOR*: 1 in place of single element with minimum signed value,
              or the one with lowest index if there is a tie.

            * *MIN_VAL*: signed value of the element with the minimum signed value,
              or all elements with the minimum value if there is a tie.

            * *MIN_ABS_VAL*: absolute value of element with the minimum absolute value;
              if there is a tie, which elements are returned is determined by `tie_index <OneHot.tie_index>`.

            * *MIN_INDICATOR*: 1 in place of the element with the minimum signed value;
              if there is a tie, which elements are returned is determined by `tie_index <OneHot.tie_index>`.

            * *MIN_ABS_INDICATOR*: 1 in place of the element with the minimum absolute value;
              if there is a tie, which elements are returned is determined by `tie_index <OneHot.tie_index>`.

    Arguments
    ---------

    variable : 2d np.array : default class_defaults.variable
        First (possibly only) item specifies a template for the array to be transformed;  if `mode <OneHot.mode>` is
        *PROB* then a 2nd item must be included that is a probability distribution with same length as 1st item.

    mode : DETERMINISTiC, PROB, PROB_INDICATOR,
    ARG_MAX, ARG_MAX_ABS, ARG_MAX_INDICATOR, ARG_MAX_ABS_INDICATOR,
    ARG_MIN, ARG_MIN_ABS, ARG_MIN_INDICATOR, ARG_MIN_ABS_INDICATOR,
    MAX_VAL, MAX_ABS_VAL, MAX_INDICATOR, MAX_ABS_INDICATOR,
    MIN_VAL, MIN_ABS_VAL, MIN_INDICATOR,  MIN_ABS_INDICATOR,
    : default ARG_MAX
        specifies how non-zero value(s) in the array returned by `function <OneHot.function>` are determined
        (see `above <OneHot>` for details).

    direction : MAX or MIN : default MAX
       specifies whether the maximum or minimum value(s) in the array are selected.
       (see `above <OneHot>` for details).

    abs_val : bool : default False
       specifies whether the absolute values of the elements in the array are used to
       select the maximum or minimum value(s).
       (see `above <OneHot>` for details).

    indicator :  bool : default False
       specifies whether the selected values(s) is (are) replace with a value of 1.
       (see `above <OneHot>` for details).

    tie : ALL, FIRST, LAST, RANDOM : default ALL
       specifies how a tie is handled when there is more than one element with the maximum or minimum value;
       (see `above <OneHot>` for details).

    params : Dict[param keyword: param value] : default None
        a `parameter dictionary <ParameterPort_Specification>` that specifies the parameters for the
        function.  Values specified for parameters in the dictionary override any assigned to those parameters in
        arguments of the constructor.

    bounds : None

    owner : Component
        `component <Component>` to which to assign the Function.

    name : str : default see `name <Function.name>`
        specifies the name of the Function.

    prefs : PreferenceSet or specification dict : default Function.classPreferences
        specifies the `PreferenceSet` for the Function (see `prefs <Function_Base.prefs>` for details).

    Attributes
    ----------

    variable : number or np.array
        1st item contains value to be transformed;  if `mode <OneHot.mode>` is *PROB*, 2nd item is a probability
        distribution, each element of which specifies the probability for selecting the corresponding element of the
        1st item.

    mode : DETERMINISTIC, PROB, PROB_INDICATOR,
    ARG_MAX, ARG_MAX_ABS, ARG_MAX_INDICATOR, ARG_MAX_ABS_INDICATOR,
    ARG_MIN, ARG_MIN_ABS, ARG_MIN_INDICATOR, ARG_MIN_ABS_INDICATOR,
    MAX_VAL, MAX_ABS_VAL, MAX_INDICATOR, MAX_ABS_INDICATOR,
    MIN_VAL, MIN_ABS_VAL, MIN_INDICATOR,  MIN_ABS_INDICATOR,
    : default ARG_MAX
        specifies how non-zero value(s) in the array returned by `function <OneHot.function>` are determined
        (see `above <OneHot>` for details).

    direction : MAX or MIN
       determines whether the maximum or minimum value(s) in the array are selected.
       (see `above <OneHot>` for details).

    abs_val : bool
       determines whether the absolute values of the elements in the array are used to
       select the maximum or minimum value(s).
       (see `above <OneHot>` for details).

    indicator :  bool
       determines whether the selected values(s) is (are) replace with a value of 1.
       (see `above <OneHot>` for details).

    tie : ALL, FIRST, LAST, RANDOM
       determines how a tie is handled when there is more than one element with the maximum or minimum value;
       (see `above <OneHot>` for details).

    random_state : numpy.RandomState
        private pseudorandom number generator

    owner : Component
        `component <Component>` to which the Function has been assigned.

    name : str
        the name of the Function; if it is not specified in the **name** argument of the constructor, a default is
        assigned by FunctionRegistry (see `Registry_Naming` for conventions used for default and duplicate names).

    prefs : PreferenceSet or specification dict : Function.classPreferences
        the `PreferenceSet` for function; if it is not specified in the **prefs** argument of the Function's
        constructor, a default is assigned using `classPreferences` defined in __init__.py (see `Preferences`
        for details).
    """

    componentName = ONE_HOT_FUNCTION

    classPreferences = {
        PREFERENCE_SET_NAME: 'OneHotClassPreferences',
        REPORT_OUTPUT_PREF: PreferenceEntry(False, PreferenceLevel.INSTANCE),
    }

    class Parameters(SelectionFunction.Parameters):
        """
            Attributes
            ----------

                mode
                    see `mode <OneHot.mode>`

                    :default value: `DETERMINISTIC`
                    :type: ``str``

                direction
                    see `direction <OneHot.direction>`

                    :default value: `MAX`
                    :type: ``str``

                abs_val
                    see `abs_val <OneHot.abs_val>`

                    :default value: `False`
                    :type: ``bool``

                indicator
                    see `indicator <OneHot.indicator>`

                    :default value: `False`
                    :type: ``bool``

                tie
                    see `tie <OneHot.tie>`

                    :default value: `ALL`
                    :type: ``str``

                random_state
                    see `random_state <OneHot.random_state>`

                    :default value: None
                    :type: ``numpy.random.RandomState``
        """
        mode = Parameter(DETERMINISTIC, stateful=False)
        direction = Parameter(MAX, stateful=False)
        abs_val = Parameter(False, stateful=False)
        indicator = Parameter(False, stateful=False)
        tie = Parameter(ALL, stateful=False)
        random_state = Parameter(None, loggable=False, getter=_random_state_getter, dependencies='seed')
        seed = Parameter(DEFAULT_SEED(), modulable=True, fallback_default=True, setter=_seed_setter)

        def _validate_mode(self, mode):
            if mode not in mode_options:
                # returns error message
                return 'not one of {0}'.format(mode_options)

        def _validate_ties(self, tie_index):
            if tie_index not in tie_options:
                # returns error message
                return 'not one of {0}'.format(tie_options)

    @check_user_specified
    @beartype
    def __init__(self,
                 default_variable=None,
                 mode: Optional[Literal[
                     DETERMINISTIC, PROB, PROB_INDICATOR,
                     ARG_MAX, ARG_MAX_ABS, ARG_MAX_INDICATOR, ARG_MAX_ABS_INDICATOR,
                     ARG_MIN, ARG_MIN_ABS, ARG_MIN_INDICATOR, ARG_MIN_ABS_INDICATOR,
                     MAX_VAL, MAX_ABS_VAL, MAX_INDICATOR, MAX_ABS_INDICATOR,
                     MIN_VAL, MIN_ABS_VAL, MIN_INDICATOR, MIN_ABS_INDICATOR]] = None,
                 direction: Optional[Literal[MAX, MIN]] = None,
                 abs_val: Optional[bool] = None,
                 indicator: Optional[bool] = None,
                 tie: Optional[Literal[ALL, FIRST, LAST, RANDOM]]= None,
                 seed=None,
                 params=None,
                 owner=None,
                 prefs:  Optional[ValidPrefSet] = None):

        reset_variable_shape_flexibility = False
        if mode in {PROB, PROB_INDICATOR} and default_variable is None:
            default_variable = [[0], [0]]
            reset_variable_shape_flexibility = True

        super().__init__(
            default_variable=default_variable,
            mode=mode,
            direction=direction,
            abs_val=abs_val,
            indicator=indicator,
            tie=tie,
            seed=seed,
            params=params,
            owner=owner,
            prefs=prefs,
        )

        if reset_variable_shape_flexibility:
            self._variable_shape_flexibility = DefaultsFlexibility.FLEXIBLE

    def _validate_params(self, request_set, target_set=None, context=None):

        if request_set[MODE] in {PROB, PROB_INDICATOR}:
            if not self.defaults.variable.ndim == 2:
                raise FunctionError("If {} for {} {} is set to {}, variable must be 2d array".
                                    format(MODE, self.__class__.__name__, Function.__name__, PROB))
            values = self.defaults.variable[0]
            prob_dist = self.defaults.variable[1]
            if len(values)!=len(prob_dist):
                raise FunctionError("If {} for {} {} is set to {}, the two items of its variable must be of equal "
                                    "length (len item 1 = {}; len item 2 = {}".
                                    format(MODE, self.__class__.__name__, Function.__name__, PROB,
                                           len(values), len(prob_dist)))
            if not all((elem>=0 and elem<=1) for elem in prob_dist)==1:
                raise FunctionError("If {} for {} {} is set to {}, the 2nd item of its variable ({}) must be an "
                                    "array of elements each of which is in the (0,1) interval".
                                    format(MODE, self.__class__.__name__, Function.__name__, PROB, prob_dist))
            if self.is_initializing:
                return
            # FIX 8/20/23: WHY DOES SUM COME UP WITH FLOATING POINT ERRORS?
            # if not np.sum(prob_dist)==1:
            if not np.allclose(np.sum(prob_dist), 1):
                raise FunctionError("If {} for {} {} is set to {}, the 2nd item of its variable ({}) must be an "
                                    "array of probabilities that sum to 1".
                                    format(MODE, self.__class__.__name__, Function.__name__, PROB, prob_dist))

        elif request_set[MODE] != DETERMINISTIC:
            # Ensure that mode is not specified counter to other options (except tie)
            if any([self.parameters.direction._user_specified,
                   self.parameters.abs_val._user_specified,
                   self.parameters.indicator._user_specified]):
                raise FunctionError(f"If {MODE} for {self.__class__.__name__} {Function.__name__} is not "
                                    f"set to 'DETERMINIST', then the 'direction', 'abs_val', and 'indicator' args "
                                    f"cannot be specified.")

    def _gen_llvm_function_body(self, ctx, builder, params, state, arg_in, arg_out, *, tags:frozenset):
        if self.mode in {PROB, PROB_INDICATOR}:

            sum_ptr = builder.alloca(ctx.float_ty)
            builder.store(sum_ptr.type.pointee(-0.0), sum_ptr)

            rand_state_ptr = ctx.get_random_state_ptr(builder, self, state, params)
            rng_f = ctx.get_uniform_dist_function_by_state(rand_state_ptr)
            random_draw_ptr = builder.alloca(rng_f.args[-1].type.pointee)
            builder.call(rng_f, [rand_state_ptr, random_draw_ptr])
            random_draw = builder.load(random_draw_ptr)

            prob_in = builder.gep(arg_in, [ctx.int32_ty(0), ctx.int32_ty(1)])
            arg_in = builder.gep(arg_in, [ctx.int32_ty(0), ctx.int32_ty(0)])

            with pnlvm.helpers.array_ptr_loop(builder, arg_in, "search") as (b1, idx):

                current_ptr = b1.gep(arg_in, [ctx.int32_ty(0), idx])
                current = b1.load(current_ptr)

                # Update prefix sum
                current_prob_ptr = b1.gep(prob_in, [ctx.int32_ty(0), idx])
                sum_old = b1.load(sum_ptr)
                sum_new = b1.fadd(sum_old, b1.load(current_prob_ptr))
                b1.store(sum_new, sum_ptr)

                old_below = b1.fcmp_ordered("<=", sum_old, random_draw)
                new_above = b1.fcmp_ordered("<", random_draw, sum_new)
                cond = b1.and_(new_above, old_below)

                if self.mode == PROB:
                    val = current
                else:
                    val = current.type(1.0)

                write_val = b1.select(cond, val, val.type(0.0))
                cur_res_ptr = b1.gep(arg_out, [ctx.int32_ty(0), idx])
                builder.store(write_val, cur_res_ptr)

            return builder

        elif self.mode == DETERMINISTIC:
            direction = self.direction
            tie = self.tie
            abs_val_ptr = ctx.get_param_or_state_ptr(builder, self, self.parameters.abs_val, param_struct_ptr=params)
            indicator_ptr = ctx.get_param_or_state_ptr(builder, self, self.parameters.indicator, param_struct_ptr=params)

            abs_val = builder.load(abs_val_ptr)
            is_abs_val = builder.fcmp_unordered("!=", abs_val, abs_val.type(0))

            indicator = builder.load(indicator_ptr)
            is_indicator = builder.fcmp_unordered("!=", indicator, indicator.type(0))

        else:
            direction, abs_val, indicator, tie = self._parse_mode(self.mode)
            is_abs_val = ctx.bool_ty(abs_val)
            is_indicator = ctx.bool_ty(indicator)

        num_extremes_ptr = builder.alloca(ctx.int32_ty)
        builder.store(num_extremes_ptr.type.pointee(0), num_extremes_ptr)

        extreme_val_ptr = builder.alloca(ctx.float_ty)
        builder.store(extreme_val_ptr.type.pointee(float("NaN")), extreme_val_ptr)

        fabs_f = ctx.get_builtin("fabs", [extreme_val_ptr.type.pointee])

        with pnlvm.helpers.recursive_iterate_arrays(ctx, builder, arg_in, loop_id="count_extremes") as (loop_builder, current_ptr):

            current = loop_builder.load(current_ptr)
            current_abs = loop_builder.call(fabs_f, [current])
            current = builder.select(is_abs_val, current_abs, current)

            old_extreme = loop_builder.load(extreme_val_ptr)
            cmp_op = ">" if direction == MAX else "<"
            is_new_extreme = loop_builder.fcmp_unordered(cmp_op, current, old_extreme)

            with loop_builder.if_then(is_new_extreme):
                loop_builder.store(current, extreme_val_ptr)
                loop_builder.store(num_extremes_ptr.type.pointee(1), num_extremes_ptr)

            is_old_extreme = loop_builder.fcmp_ordered("==", current, old_extreme)
            with loop_builder.if_then(is_old_extreme):
                extreme_count = loop_builder.load(num_extremes_ptr)
                extreme_count = loop_builder.add(extreme_count, extreme_count.type(1))
                loop_builder.store(extreme_count, num_extremes_ptr)


        if tie == FIRST:
            extreme_start = num_extremes_ptr.type.pointee(0)
            extreme_stop = num_extremes_ptr.type.pointee(1)

        elif tie == LAST:
            extreme_stop = builder.load(num_extremes_ptr)
            extreme_start = builder.sub(extreme_stop, extreme_stop.type(1))

        elif tie == ALL:
            extreme_start = num_extremes_ptr.type.pointee(0)
            extreme_stop = builder.load(num_extremes_ptr)

        elif tie == RANDOM:
            rand_state_ptr = ctx.get_random_state_ptr(builder, self, state, params)
            rand_f = ctx.get_rand_int_function_by_state(rand_state_ptr)
            random_draw_ptr = builder.alloca(rand_f.args[-1].type.pointee)
            num_extremes = builder.load(num_extremes_ptr)

            builder.call(rand_f, [rand_state_ptr, ctx.int32_ty(0), num_extremes, random_draw_ptr])

            extreme_start = builder.load(random_draw_ptr)
            extreme_start = builder.trunc(extreme_start, ctx.int32_ty)
            extreme_stop = builder.add(extreme_start, extreme_start.type(1))

        else:
            assert False, "Unknown tie resolution: {}".format(tie)


        extreme_val = builder.load(extreme_val_ptr)
        extreme_write_val = builder.select(is_indicator, extreme_val.type(1), extreme_val)
        next_extreme_ptr = builder.alloca(num_extremes_ptr.type.pointee)
        builder.store(next_extreme_ptr.type.pointee(0), next_extreme_ptr)

        pnlvm.helpers.printf(ctx,
                             builder,
                             "{} replacing extreme values of %e from <%u,%u) out of %u\n".format(self.name),
                             extreme_val,
                             extreme_start,
                             extreme_stop,
                             builder.load(num_extremes_ptr),
                             tags={"one_hot"})

        with pnlvm.helpers.recursive_iterate_arrays(ctx, builder, arg_in, arg_out, loop_id="mark_extremes") as (loop_builder, current_ptr, out_ptr):
            current = loop_builder.load(current_ptr)
            current_abs = loop_builder.call(fabs_f, [current])
            current = builder.select(is_abs_val, current_abs, current)

            is_extreme = loop_builder.fcmp_ordered("==", current, extreme_val)
            current_extreme_idx = loop_builder.load(next_extreme_ptr)

            with loop_builder.if_then(is_extreme):
                next_extreme_idx = loop_builder.add(current_extreme_idx, current_extreme_idx.type(1))
                loop_builder.store(next_extreme_idx, next_extreme_ptr)

            is_after_start = loop_builder.icmp_unsigned(">=", current_extreme_idx, extreme_start)
            is_before_stop = loop_builder.icmp_unsigned("<", current_extreme_idx, extreme_stop)

            should_write_extreme = loop_builder.and_(is_extreme, is_after_start)
            should_write_extreme = loop_builder.and_(should_write_extreme, is_before_stop)

            write_value = loop_builder.select(should_write_extreme, extreme_write_val, extreme_write_val.type(0))
            loop_builder.store(write_value, out_ptr)

        return builder

    def _parse_mode(self, mode):
        """Convert mode spec to corresponding options.
        Here for convenience, but mostly for backward compatibility with old mode spec.
        """

        direction = None
        abs_val = None
        indicator = None
        tie = None

        if mode == ARG_MAX:
            direction = MAX
            abs_val = False
            indicator = False
            tie = FIRST

        elif mode == ARG_MAX_ABS:
            direction = MAX
            abs_val = True
            indicator = False
            tie = FIRST

        elif mode == ARG_MAX_INDICATOR:
            direction = MAX
            abs_val = False
            indicator = True
            tie = FIRST

        elif mode == ARG_MAX_ABS_INDICATOR:
            direction = MAX
            abs_val = True
            indicator = True
            tie = FIRST

        elif mode == MAX_VAL:
            direction = MAX
            abs_val = False
            indicator = False
            tie = ALL

        elif mode == MAX_ABS_VAL:
            direction = MAX
            abs_val = True
            indicator = False
            tie = ALL

        elif mode == MAX_INDICATOR:
            direction = MAX
            abs_val = False
            indicator = True
            tie = ALL

        elif mode == MAX_ABS_INDICATOR:
            direction = MAX
            abs_val = True
            indicator = True
            tie = ALL

        elif mode == ARG_MIN:
            direction = MIN
            abs_val = False
            indicator = False
            tie = FIRST

        elif mode == ARG_MIN_ABS:
            direction = MIN
            abs_val = True
            indicator = False
            tie = FIRST

        elif mode == ARG_MIN_INDICATOR:
            direction = MIN
            abs_val = False
            indicator = True
            tie = FIRST

        elif mode == ARG_MIN_ABS_INDICATOR:
            direction = MIN
            abs_val = True
            indicator = True
            tie = FIRST

        elif mode == MIN_VAL:
            direction = MIN
            abs_val = False
            indicator = False
            tie = ALL

        elif mode == MIN_ABS_VAL:
            direction = MIN
            abs_val = True
            indicator = False
            tie = ALL

        elif mode == MIN_INDICATOR:
            direction = MIN
            abs_val = False
            indicator = True
            tie = ALL

        elif mode == MIN_ABS_INDICATOR:
            direction = MIN
            abs_val = True
            indicator = True
            tie = ALL

        else:
            assert False, f"Unknown mode: {mode}"

        return direction, abs_val, indicator, tie

    def _function(self,
                 variable=None,
                 context=None,
                 params=None,
                 ):
        """

        Arguments
        ---------

        variable : 2d np.array : default class_defaults.variable
           1st item is an array to be transformed;  if `mode <OneHot.mode>` is *PROB*, 2nd item must be an array of
           probabilities (i.e., elements between 0 and 1) of equal length as the 1st item.

        params : Dict[param keyword: param value] : default None
            a `parameter dictionary <ParameterPort_Specification>` that specifies the parameters for the
            function.  Values specified for parameters in the dictionary override any assigned to those parameters in
            arguments of the constructor.

        Returns
        -------

        array with selected elements having non-zero values and all others having zeroes : np.array
            specified by `mode <OneHot.mode>`, `direction <OneHot.direction>`, `abs_val <OneHot.abs_val>`,
            `indicator <OneHot.indicator>`, and `tie <OneHot.tie>`.

        """

        mode = self.parameters.mode.get(context)
        direction = self.parameters.direction.get(context)
        abs_val = self.parameters.abs_val.get(context)
        indicator = self.parameters.indicator.get(context)
        tie = self.parameters.tie.get(context)

        if mode in {PROB, PROB_INDICATOR}:
            # 1st item of variable should be data, and 2nd a probability distribution for choosing
            if np.array(variable).ndim != 2:
                raise FunctionError(f"If {MODE} for {self.__class__.__name__} {Function.__name__} is set to "
                                    f"'PROB' or 'PROB_INDICATOR', variable must be a 2d array with the first item "
                                    f"being the data and the second being a probability distribution.")
            v = variable[0]
            prob_dist = variable[1]
            # if not prob_dist.any() and INITIALIZING in context:
            if not prob_dist.any():
                return self.convert_output_type(v)
            cum_sum = np.cumsum(prob_dist)
            random_state = self._get_current_parameter_value("random_state", context)
            random_value = random_state.uniform()
            chosen_item = next(element for element in cum_sum if element > random_value)
            chosen_in_cum_sum = np.where(cum_sum == chosen_item, 1, 0)
            if mode == PROB:
                result = v * chosen_in_cum_sum
            else:
                result = np.ones_like(v) * chosen_in_cum_sum

            # chosen_item = np.random.choice(v, 1, p=prob_dist)
            # one_hot_indicator = np.where(v == chosen_item, 1, 0)
            # return v * one_hot_indicator
            return result

        elif mode != DETERMINISTIC:
            direction, abs_val, indicator, tie = self._parse_mode(mode)

        array = variable

        if abs_val is True:
            array = np.absolute(variable)

        if direction == MAX:
            extreme_val = np.max(array)
            if extreme_val == -np.inf:
                warnings.warn(f"Array passed to {self.name} of {self.owner.name} is all -inf.")

        elif direction == MIN:
            extreme_val = np.min(array)
            if extreme_val == np.inf:
                warnings.warn(f"Array passed to {self.name} of {self.owner.name} is all inf.")

        else:
            assert False, f"Unknown direction: '{direction}'."

        extreme_indices = np.where(array == extreme_val)

        num_indices = len(extreme_indices[0])
        assert all(len(idx) == num_indices for idx in extreme_indices)

        if tie == FIRST:
            selected_idx = 0

        elif tie == LAST:
            selected_idx = -1

        elif tie == RANDOM:
            random_state = self._get_current_parameter_value("random_state", context)
            selected_idx = random_state.randint(num_indices)

        elif tie == ALL:
            selected_idx = slice(num_indices)

        else:
            assert False, f"PROGRAM ERROR: Unrecognized value for 'tie' in OneHot function: '{tie}'."


        set_indices = tuple(index[selected_idx] for index in extreme_indices)

        result = np.zeros_like(variable)
        result[set_indices] = 1 if indicator else extreme_val

        return self.convert_output_type(result)



#%%
import time
import torch
from torch import nn

from typing import Tuple

if torch.cuda.is_available():
  dev = "cuda:0"
else:
  dev = "cpu"


class DriftDiffusionModel(torch.nn.Module):
    def forward(self,
                starting_value: float = 0.0,
                rate: float = 1.0,
                non_decision_time: float = 0.0,
                threshold: float = 1.0,
                noise: float = 1.0,
                time_step_size: float = 0.01,
                num_walkers: int = 1000,
                dev: str = "cuda:0") -> Tuple[torch.Tensor, torch.Tensor]:

        """
        A model that simulates many instances of a simple noisy drift diffusion model in parallel.

        Args:
            starting_value: The starting value of each particle in the model.
            rate: The drift rate for each particle.
            non_decision_time: A constant amount of time added to each reaction time that signifies automatic processing
                times of stimuli.
            threshold: The threshold that a particle must reach to stop integration.
            noise: The standard deviation of the Gaussian noise added to each particles position at each time step.
            time_step_size: The time step size (in seconds) for the integration process.
            num_walkers: The number of particles to simulate.
            dev: The device the model should be run on.

        Returns:
            A two element tuple containing the reaction times and the decisions
        """

        particle = torch.ones(size=(num_walkers,), device=dev) * starting_value
        active = torch.ones(size=(num_walkers,), dtype=torch.bool, device=dev)
        rts = torch.zeros(size=(num_walkers,), device=dev)

        for i in range(3000):
            #dw = torch.distributions.Normal(loc=rate * time_step_size * active, scale=noise * active).rsample()
            dw = torch.normal(mean=rate * time_step_size * active, std=noise * active)
            particle = particle + dw * torch.sqrt(time_step_size)
            rts = rts + active
            active = torch.abs(particle) < threshold

        rts = (non_decision_time + rts * time_step_size)

        decisions = torch.ones(size=(num_walkers,), device=dev)
        decisions[particle <= -threshold] = 0

        return rts, decisions


ddm_params = dict(starting_value=0.0, rate=0.3, non_decision_time=0.15, threshold=0.6, noise=1.0, time_step_size=0.001)

NUM_WALKERS = 1000000

# Move params to device
for key, val in ddm_params.items():
    ddm_params[key] = torch.tensor(val).to(dev)

#%%

t0 = time.time()
ddm_model = DriftDiffusionModel()
rts, decision = ddm_model(**ddm_params, num_walkers=NUM_WALKERS, dev=dev)
rts = rts.to("cpu")
decision = decision.to("cpu")
print(f"PyTorch Elapsed: {1000.0 * (time.time() - t0)} milliseconds")

#%%

# JIT
jit_ddm_model = torch.jit.script(ddm_model)

#%%
# rts, decision = jit_ddm_model(**ddm_params, num_walkers=NUM_WALKERS)
#
# NUM_TIMES = 50
# t0 = time.time()
# for i in range(NUM_TIMES):
#     rts, decision = jit_ddm_model(**ddm_params, num_walkers=NUM_WALKERS)
#     rts = rts.to("cpu")
#     decision = decision.to("cpu")
# print(f"JIT Elapsed: {1000 * ((time.time() - t0) / NUM_TIMES)} milliseconds")

#%%

jit_ddm_model.save("ddm.pt")

#%%
# with open('ddm.onnx', 'wb') as file:
#     torch.onnx.export(model=torch.jit.script(DriftDiffusionModel()),
#                       args=tuple(ddm_params.values()) + (torch.tensor(NUM_WALKERS),),
#                       example_outputs=(rts, decision),
#                       f=file,
#                       verbose=True,
#                       opset_version=12)

#%%
# import seaborn as sns
# sns.kdeplot(rts)



import timeit

import numpy as np
from psyneulink import *

from gym_forager.envs.forager_env import ForagerEnv
# Runtime Switches:
RENDER = False

# *********************************************************************************************************************
# *********************************************** CONSTANTS ***********************************************************
# *********************************************************************************************************************

# These should probably be replaced by reference to ForagerEnv constants:
obs_len = 3
obs_coords = 2
action_len = 2
player_idx = 0
player_obs_start_idx = player_idx * obs_len
player_value_idx = player_idx * obs_len + obs_coords
player_coord_slice = slice(player_obs_start_idx,player_value_idx)
predator_idx = 1
predator_obs_start_idx = predator_idx * obs_len
predator_value_idx = predator_idx * obs_len + obs_coords
predator_coord_slice = slice(predator_obs_start_idx,predator_value_idx)
prey_idx = 2
prey_obs_start_idx = prey_idx * obs_len
prey_value_idx = prey_idx * obs_len + obs_coords
prey_coord_slice = slice(prey_obs_start_idx,prey_value_idx)

player_len = prey_len = predator_len = obs_coords

dist = Distance(metric=EUCLIDEAN)

PREDATOR = 0
PREY = 1
ATTEND = 0
DISATTEND = 500
UNDECIDED = 0


# *********************************************************************************************************************
# *********************************  SAPMLE FUNCTIONS USED FOR CONTROL  ***********************************************
# *********************************************************************************************************************
# NOTE:  THESE WILL BE REPLACED BY OCM-SPECIFIC FUNCTIONS (AS DECRIBED BELOW) WHEN THAT IS IMPLEMENTED


def choose_closer_agent_function(variable):
    if variable is None:
        return [0,0]
    player_coord = variable[0]
    predator_coord = variable[1]
    prey_coord = variable[2]
    dist_to_predator = dist([player_coord, predator_coord])
    dist_to_prey = dist([player_coord, prey_coord])
    denom = dist_to_predator + dist_to_prey
    if denom != 0:
        normalized_dist_to_predator = dist_to_predator / denom
        normalized_dist_to_prey = dist_to_prey / denom
        if normalized_dist_to_predator < normalized_dist_to_prey:
            return [PREDATOR]
        else:
            return [PREY]
    return [-1]

def control_allocation_function(variable):
    closest_agent = variable[0]

    if closest_agent == PREDATOR:
        return [[ATTEND],[DISATTEND]]
    elif closest_agent == PREY:
        return [[DISATTEND],[ATTEND]]
    else:
        return [[UNDECIDED],[UNDECIDED]]

# *********************************************************************************************************************
# **************************************  MECHANISMS AND COMPOSITION  *************************************************
# *********************************************************************************************************************

# Perceptual Mechanisms
player_obs = ProcessingMechanism(input_shapes=prey_len, function=GaussianDistort, name="PLAYER OBS")
prey_obs = ProcessingMechanism(input_shapes=prey_len, function=GaussianDistort, name="PREY OBS")
predator_obs = TransferMechanism(input_shapes=predator_len, function=GaussianDistort, name="PREDATOR OBS")

# Value and Reward Mechanisms (not yet used;  for future use)
values = TransferMechanism(input_shapes=3, name="AGENT VALUES")
reward = TransferMechanism(name="REWARD")

# Action Mechanism
#    Use ComparatorMechanism to compute direction of action as difference of coordinates between player and prey:
#    note: unitization is done in main loop, to allow compilation of LinearCombination function) (TBI)
greedy_action_mech = ComparatorMechanism(name='ACTION',sample=player_obs,target=prey_obs)

# ControlMechanism
Panicky_control_mech = ControlMechanism(objective_mechanism=ObjectiveMechanism(function=choose_closer_agent_function,
                                                                               monitored_output_ports=[player_obs,
                                                                                                        predator_obs,
                                                                                                        prey_obs]),
                                        function = control_allocation_function,
                                        control_signals=[(VARIANCE,predator_obs), (VARIANCE,prey_obs)]
)

# Create Composition
agent_comp = Composition(name='PANICKY CONTROL COMPOSITION')
agent_comp.add_node(player_obs)
agent_comp.add_node(prey_obs)
agent_comp.add_node(predator_obs)
agent_comp.add_node(greedy_action_mech)
agent_comp.add_node((Panicky_control_mech))

# agent_comp.show_graph()


# *********************************************************************************************************************
# ******************************************   RUN SIMULATION  ********************************************************
# *********************************************************************************************************************

num_trials = 4

def main():

    env = ForagerEnv()
    reward = 0
    done = False
    if RENDER:
        env.render()  # If visualization is desired
    else:
        print("Running simulation...")
    steps = 0
    start_time = timeit.default_timer()
    for _ in range(num_trials):
        observation = env.reset()
        while True:
            run_results = agent_comp.run(inputs={
                player_obs:[observation[player_coord_slice]],
                predator_obs:[observation[predator_coord_slice]],
                prey_obs:[observation[prey_coord_slice]],
                Panicky_control_mech:[0]
                # values:[observation[player_value_idx],observation[prey_value_idx],observation[predator_value_idx]],
                # reward:[reward],
            })
            action= np.where(run_results[0] == 0, 0, run_results[0] / np.abs(run_results[0]))
            observation, reward, done, _ = env.step(action)
            steps +=1
            if done:
                break
    stop_time = timeit.default_timer()
    print(f'{steps / (stop_time - start_time):.1f} steps/second, {steps} total steps in '
          f'{stop_time - start_time:.2f} seconds')
    if RENDER:
        env.render()  # If visualization is desired

if __name__ == "__main__":
    main()



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# ********************************************* Condition **************************************************************

import collections
import inspect
import numbers
import warnings

import dill
import graph_scheduler
import numpy as np

from psyneulink.core.globals.context import handle_external_context
from psyneulink.core.globals.mdf import MDFSerializable
from psyneulink.core.globals.keywords import MODEL_SPEC_ID_TYPE, comparison_operators
from psyneulink.core.globals.parameters import parse_context
from psyneulink.core.globals.utilities import parse_valid_identifier, toposort_key


__all__ = [  # noqa: F822  (dynamically generated)
    'AbsoluteCondition', 'AddEdgeTo', 'AfterCall', 'AfterNCalls',
    'AfterNCallsCombined', 'AfterNode', 'AfterNodes', 'AfterNPasses',
    'AfterNRuns', 'AfterNTimeSteps', 'AfterNTrials', 'AfterPass',
    'AfterRun', 'AfterTimeStep', 'AfterTrial', 'All', 'AllHaveRun',
    'Always', 'And', 'Any', 'AtNCalls', 'AtPass', 'AtRun',
    'AtRunNStart', 'AtRunStart', 'AtTimeStep', 'AtTrial',
    'AtTrialNStart', 'AtTrialStart', 'BeforeNCalls', 'BeforeNode',
    'BeforeNodes', 'BeforePass', 'BeforeTimeStep', 'BeforeTrial',
    'CompositeCondition', 'Condition', 'ConditionBase',
    'ConditionError', 'ConditionSet', 'CustomGraphStructureCondition',
    'EveryNCalls', 'EveryNPasses', 'GraphStructureCondition', 'JustRan',
    'Never', 'Not', 'NWhen', 'Operation', 'Or', 'RemoveEdgeFrom',
    'Threshold', 'TimeInterval', 'TimeTermination', 'When',
    'WhenFinished', 'WhenFinishedAll', 'WhenFinishedAny', 'While',
    'WhileNot', 'WithNode',
]


# avoid restricting graph_scheduler versions for this code
# ConditionBase was introduced with graph structure conditions
gs_condition_base_class = graph_scheduler.condition.Condition
condition_class_parents = [graph_scheduler.condition.Condition]


try:
    gs_condition_base_class = graph_scheduler.condition.ConditionBase
except AttributeError:
    pass
else:
    class ConditionBase(graph_scheduler.condition.ConditionBase, MDFSerializable):
        def as_mdf_model(self):
            raise graph_scheduler.ConditionError(
                f'MDF support not yet implemented for {type(self)}'
            )
    condition_class_parents.append(ConditionBase)


try:
    graph_scheduler.condition.GraphStructureCondition
except AttributeError:
    graph_structure_conditions_available = False
    gsc_unavailable_message = (
        'Graph structure conditions are not available'
        f'in your installed graph-scheduler v{graph_scheduler.__version__}'
    )
else:
    graph_structure_conditions_available = True
    gsc_unavailable_message = ''


def _create_as_pnl_condition(condition):
    import psyneulink as pnl

    try:
        pnl_class = getattr(pnl.core.scheduling.condition, type(condition).__name__)
    except (AttributeError, TypeError):
        return condition

    # already a pnl Condition
    if isinstance(condition, pnl_condition_base_class):
        return condition

    if not issubclass(pnl_class, gs_condition_base_class):
        return None

    if (
        graph_structure_conditions_available
        and isinstance(condition, graph_scheduler.condition.GraphStructureCondition)
    ):
        try:
            return pnl_class(
                *condition.nodes,
                **{k: v for k, v in condition.kwargs.items() if k != 'nodes'}
            )
        except AttributeError:
            return pnl_class(**condition.kwargs)

    new_args = [_create_as_pnl_condition(a) or a for a in condition.args]
    new_kwargs = {k: _create_as_pnl_condition(v) or v for k, v in condition.kwargs.items()}
    sig = inspect.signature(pnl_class)

    if 'func' in sig.parameters or 'function' in sig.parameters:
        # Condition takes a function as an argument
        res = pnl_class(condition.func, *new_args, **new_kwargs)
    else:
        res = pnl_class(*new_args, **new_kwargs)

    res.owner = condition.owner
    return res


class Condition(*condition_class_parents, MDFSerializable):
    @handle_external_context()
    def is_satisfied(self, *args, context=None, execution_id=None, **kwargs):
        if execution_id is None:
            try:
                execution_id = parse_context(context).execution_id
            except AttributeError:
                pass

        return super().is_satisfied(
            *args,
            context=context,
            execution_id=execution_id,
            **kwargs
        )

    def as_mdf_model(self):
        import modeci_mdf.mdf as mdf
        from psyneulink.core.components.component import Component

        def _parse_condition_arg(arg):
            if isinstance(arg, Component):
                return parse_valid_identifier(arg.name)
            elif isinstance(arg, Condition):
                return arg.as_mdf_model()
            elif (
                isinstance(arg, np.number)
                or (isinstance(arg, np.ndarray) and arg.ndim == 0)
            ):
                return arg.item()
            elif arg is None or isinstance(arg, numbers.Number):
                return arg
            else:
                try:
                    iter(arg)
                except TypeError:
                    return str(arg)
                else:
                    return arg

        if type(self) in {graph_scheduler.Condition, Condition}:
            try:
                func_val = inspect.getsource(self.func)
            except OSError:
                func_val = dill.dumps(self.func)
            func_dict = {'function': func_val}
        else:
            func_dict = {}

        extra_args = {MODEL_SPEC_ID_TYPE: getattr(graph_scheduler.condition, type(self).__name__).__name__}

        sig = inspect.signature(self.__init__)

        for name, param in sig.parameters.items():
            if param.kind is inspect.Parameter.VAR_POSITIONAL:
                args_list = []
                for a in self.args:
                    if isinstance(a, Component):
                        a = parse_valid_identifier(a.name)
                    elif isinstance(a, Condition):
                        a = a.as_mdf_model()
                    args_list.append(a)
                extra_args[name] = args_list

        for i, (name, param) in enumerate(filter(
            lambda item: item[1].kind is inspect.Parameter.POSITIONAL_OR_KEYWORD and item[0] not in self.kwargs,
            sig.parameters.items()
        )):
            try:
                extra_args[name] = self.args[i]
            except IndexError:
                # was specified with keyword not as positional arg
                extra_args[name] = param.default

        return mdf.Condition(
            **func_dict,
            **{
                k: _parse_condition_arg(v)
                for k, v in (*self.kwargs.items(), *extra_args.items())
            }
        )


# below produces psyneulink versions of each Condition class so that
# they are compatible with the extra changes made in Condition above
# (the scheduler does not handle Context objects or mdf/json export)
gs_class_dependencies = {}
gs_classes_to_copy_as_pnl = []
pnl_conditions_module = locals()  # inserting into locals defines the classes
pnl_condition_base_class = pnl_conditions_module[gs_condition_base_class.__name__]

for class_name in graph_scheduler.condition.__dict__:
    cls_ = getattr(graph_scheduler.condition, class_name)
    if inspect.isclass(cls_):
        # don't substitute classes explicitly defined above
        if class_name not in pnl_conditions_module:
            if issubclass(cls_, gs_condition_base_class):
                gs_classes_to_copy_as_pnl.append(class_name)
            else:
                pnl_conditions_module[class_name] = cls_

        gs_class_dependencies[class_name] = {
            c.__name__ for c in cls_.__mro__ if c.__name__ != class_name
        }

# iterate in order such that superclass types are before subclass types
for cond_name in sorted(
    gs_classes_to_copy_as_pnl,
    key=toposort_key(gs_class_dependencies)
):
    sched_module_cond_obj = getattr(graph_scheduler.condition, cond_name)
    new_bases = []
    for cls_ in sched_module_cond_obj.__mro__:
        try:
            new_bases.append(pnl_conditions_module[cls_.__name__])
        except KeyError:
            new_bases.append(cls_)
        if cls_ is gs_condition_base_class:
            break

    new_meta = type(new_bases[0])
    if new_meta is not type:
        pnl_conditions_module[cond_name] = new_meta(
            cond_name, tuple(new_bases), {'__module__': Condition.__module__}
        )
    else:
        pnl_conditions_module[cond_name] = type(cond_name, tuple(new_bases), {})

    pnl_conditions_module[cond_name].__doc__ = sched_module_cond_obj.__doc__


_doc_subs = {
    None: [
        (
            r'def converge\(node, thresh\):',
            'def converge(node, thresh, context):'
        ),
        (
            r'node\.delta',
            'node.parameters.value.get_delta(context)'
        ),
        (
            r'the ``delta`` attribute of\na node \(which reports the change in its ``value``\)',
            "the change in a node's `value`"
        )
    ]
}


class Threshold(graph_scheduler.condition._DependencyValidation, Condition):
    """Threshold

    Attributes:

        dependency
            the node on which the Condition depends

        parameter
            the name of the parameter of **dependency** whose value is
            to be compared to **threshold**

        threshold
            the fixed value compared to the value of the **parameter**

        comparator
            the string comparison operator determining the direction or
            type of comparison of the value of the **parameter**
            relative to **threshold**

        indices
            if specified, a series of indices that reach the desired
            number given an iterable value for **parameter**

        atol
            absolute tolerance for the comparison

        rtol
            relative tolerance (to **threshold**) for the comparison

    Satisfied when:

        The comparison between the value of the **parameter** and
        **threshold** using **comparator** is true. If **comparator** is
        an equality (==, !=), the comparison will be considered equal
        within tolerances **atol** and **rtol**.

    Notes:

        The comparison must be done with scalars. If the value of
        **parameter** contains more than one item, **indices** must be
        specified.
    """

    def __init__(self, dependency, parameter, threshold, comparator, indices=None, atol=0, rtol=0):
        if comparator not in comparison_operators:
            raise graph_scheduler.ConditionError(
                f'Operator must be one of {list(comparison_operators.keys())}'
            )

        if parameter not in dependency.parameters:
            raise graph_scheduler.ConditionError(
                f'{dependency} has no {parameter} parameter'
            )

        if (atol != 0 or rtol != 0) and comparator in {'<', '<=', '>', '>='}:
            warnings.warn('Tolerances for inequality comparators are ignored')

        if (
            indices is not None
            and not isinstance(indices, graph_scheduler.TimeScale)
            and not isinstance(indices, collections.abc.Iterable)
        ):
            indices = [indices]

        def func(threshold, comparator, indices, atol, rtol, execution_id):
            param_value = getattr(self.dependency.parameters, self.parameter).get(execution_id)

            if isinstance(indices, graph_scheduler.TimeScale):
                param_value = param_value._get_by_time_scale(indices)
            elif indices is not None:
                for i in indices:
                    param_value = param_value[i]

            param_value = float(np.array(param_value).item())

            if comparator == '==':
                return np.isclose(param_value, threshold, atol=atol, rtol=rtol)
            elif comparator == '!=':
                return not np.isclose(param_value, threshold, atol=atol, rtol=rtol)
            else:
                return comparison_operators[comparator](param_value, threshold)

        super().__init__(
            func,
            dependency=dependency,
            parameter=parameter,
            threshold=threshold,
            comparator=comparator,
            indices=indices,
            atol=atol,
            rtol=rtol,
        )

    def as_mdf_model(self):
        m = super().as_mdf_model()

        if self.parameter == 'value':
            m.kwargs['parameter'] = f'{self.dependency.name}_OutputPort_0'

        return m


When = Condition



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.

# NOTES:
#  * NOW THAT NOISE AND BETA ARE PROPERTIES THAT DIRECTLY REFERERNCE integrator_function,
#      SHOULD THEY NOW BE VALIDATED ONLY THERE (AND NOT IN TransferMechanism)??
#  * ARE THOSE THE ONLY TWO integrator PARAMS THAT SHOULD BE PROPERTIES??

# ****************************************  LCAMechanism *************************************************

"""

Contents
--------

  * `LCAMechanism_Overview`
  * `LCAMechanism_Creation`
      - `LCAMechanism_Integrator_Mode`
      - `LCAMechanism_Threshold`
  * `LCAMechanism_Structure`
  * `LCAMechanism_Execution`
  * `LCAMechanism_Class_Reference`


.. _LCAMechanism_Overview:

Overview
--------

An LCAMechanism is a subclass of `RecurrentTransferMechanism` that implements a single-layered `leaky competitng
accumulator (LCA) <https://www.ncbi.nlm.nih.gov/pubmed/11488378>`_  network. By default, it uses a
`LeakyCompetingIntegrator` and a `Logistic` Function to compute the activity of the units, each of which has a
self-excitatory connection (specified by the **self_excitation** argument) and mutually inhibitory connections with
every other element (specified by the **competition** argument).  These are implemented by its `recurrent_projection
<RecurrentTransferMechanism.recurrent_projection>`, the `matrix  <MappingProjection.matrix>` of which consists of
diagnoal elements assign the value of `self_excitation <LCAMechanism.self_excitation>` off-diagonal elements assigned
the negative of the value of `competition <LCAMechanism.competition>`.

    COMMENT:
    .. math::

        \\begin{bmatrix}
            excitation    &  - competition  &  - competition  &  - competition  \
            - competition &  excitation     &  - competition  &  - competition  \
            - competition &  - competition  &  excitation     &  - competition  \
            - competition &  - competition  &  - competition  &  excitation     \
        \\end{bmatrix}
    COMMENT

When all of the following conditions are true:

- the `LCAMechanism` mechanism has two elements,
- the value of its `competition <LCAMechanism.competition>` parameter is equal to its `leak <LCAMechanism.leak>`
  parameter,
- `competition <LCAMechanism.competition>` and `leak <LCAMechanism.leak>` are of sufficient magnitude,

then the `LCAMechanism` implements a close approximation of a `DDM` Mechanism (see `Usher & McClelland, 2001;
<http://psycnet.apa.org/?&fa=main.doiLanding&doi=10.1037/0033-295X.108.3.550>`_ and `Bogacz et al (2006)
<https://www.ncbi.nlm.nih.gov/pubmed/17014301>`_).

.. _LCAMechanism_Creation:

Creating an LCAMechanism
------------------------

An LCAMechanism is created by calling its constructor.  Ordinarily, the **self-excitation** and **competion**
arguments are used to specify the values of the diagonal and off-diagonal elements of the `matrix <LCAMechanism.matrix>`
of its `recurrent_projection <RecurrentTransferMechanism.recurrent_projection>` (see `LCAMechanism_Structure` below).
However, if the **matrix** argument is specified, a warning is issued and the **self_excitation** and **competition**
arguments are ignored.

.. _LCAMechanism_Integrator_Mode:

*Integration*
~~~~~~~~~~~~~

The **noise**, **leak**, **initial_value**, and **time_step_size** arguments are used to implement the
`LeakyCompetingIntegrator` as the LCAMechanism's `integrator_function <TransferMechanism.integrator_function>`.
The **leak** argument is used to specify the `leak <LeakyCompetingIntegrator.leak>` parameter of the
`LeakyCompetingIntegrator`. This function is only used used when `integrator_mode
<TransferMechanism_Integrator_Mode>` is True (which it is by default).  If `integrator_mode
<TransferMechanism.integrator_mode>` is False, the `LeakyCompetingIntegrator` function is skipped entirely,
and all related arguments (**noise**, **leak**, **initial_value**, and **time_step_size**) have no effect.

.. _LCAMechanism_Threshold:

*Thresholding*
~~~~~~~~~~~~~~

The **threshold** and **threshold_criterion** arguments specify the conditions under which execution of the
LCAMechanism terminates if `integrator_mode <TransferMechanism_Integrator_Mode>` is True.  If **threshold** is None
(the default), then the LCAMechanism will update its `value <Mechanism_Base.value>` and the `value <OutputPort.value>`
of each `OutputPort` only once each time it is executed.  If a **threshold** is specified, then it will continue
to execute until the condition specified by **threshold_criterion** is True; this can be specified using one of the
following keywords:

  * *VALUE* --  (default) True when any element of the LCAMechanism's `value <Mechanism_Base.value>` is equal to or
    greater than the **threshold**;

  * *MAX_VS_NEXT* -- True when the element of the LCAMechanism's `value <Mechanism_Base.value>` with the highest
    values is greater than the one with the next-highest value by an amount that equals or exceeds **threshold**;

  * *MAX_VS_AVG* -- True when the element of the LCAMechanism's `value <Mechanism_Base.value>` with the highest
    values is greater than the average of the others by an amount that equals or exceeds **threshold**;

  * *CONVERGENCE* -- True when the no element of the LCAMechanism's current `value <Mechanism_Base.value>`
    differs from its value on the previous update by more than **threshold**.

.. _LCAMechanism_DDM_APPROXIMATION:

For an LCAMechanism with *exactly* two elements, *MAX_VS_NEXT* implements a close approximation of the `threshold
<DriftDiffusionIntegrator.threshold>` parameter of the `DriftDiffusionIntegrator` Function used by a `DDM` (see
`Usher & McClelland, 2001; <http://psycnet.apa.org/?&fa=main.doiLanding&doi=10.1037/0033-295X.108.3.550>`_ and
`Bogacz et al (2006) <https://www.ncbi.nlm.nih.gov/pubmed/17014301>`_). For an LCAMechanism with more than two
elements, *MAX_VS_NEXT* and *MAX_VS_AVG* implements threshold approximations with different properties
(see `McMillen & Holmes, 2006 <http://www.sciencedirect.com/science/article/pii/S0022249605000891>`_).
**CONVERGENCE** (the default for a TransferMechanism) implements a "settling" process, in which the Mechanism
stops executing when updating produces sufficiently small changes.

Note that **threshold** and **threshold_criterion** are convenience arguments, and are not associated with
similarly-named attributes.  Rather, they are used to specify the `termination_threshold
<TransferMechanism.termination_threshold>`, `termination_measure <TransferMechanism.termination_measure>`,
and `termination_comparison_op <TransferMechanism.termination_comparison_op>` attributes; these can also be
specified directly as arguments of the LCAMechanism's constructor in order to implement other termination conditions
(see `TransferMechanism <TransferMechanism_Execution_Integration_Termination>` for additional details).

COMMENT:
The default format of its `variable <LCAMechanism.variable>`, and default values of its `inhibition
<LCAMechanism.inhibition>`, `decay <RecurrentTransferMechanism.decay>` and `noise <TransferMechanism.noise>` parameters
implement an approximation of a `DDM`.
COMMENT

.. _LCAMechanism_Structure:

Structure
---------

The key state_features that disinguish an LCAMechanism from its parent class (`RecurrentTransferMechainsm`) are:

1. its default `function <Mechanism.function>` is a `Logistic` Function (rather than `Linear`);

2. its default `integrator_function <TransferMechanism.integrator_function>` is a `LeakyCompetingIntegrator`
   Function (rather than `AdaptiveIntegrator`);

3. the `matrix <LCAMechanism.matrix>` of its `recurrent_projection <RecurrentTransferMechanism.recurrent_projection>`,
   by default, has diagonal elements with uniform weights assigned the value of `self_excitation
   <LCAMechanism.self_excitation>`, and off-diagonal elements with uniform weights assigned the *negative* of the
   value of `competition <LCAMechanism.competition>`;  however, if the **matrix** argument is specified, then
   `self_excitation <LCAMechanism.self_excitation>` and `competition <LCAMechanism.competition>` are ignored.

Like any RecurrentTransferMechanism, by default an LCAMechanism has a single `primary OutputPort <OutputPort_Primary>`
named *RESULT* that contains the Mechanism's current `value <Mechanism_Base.value>`.  It also has two
`StandardOutputPorts <OutputPort_Standard>` in its `standard_output_ports <LCAMechanism.standard_output_ports>`
attribute -- *MAX_VS_NEXT* and *MAX_VS_AVG* that are available for assignment, in addition to the
`standard_output_ports <RecurrentTransferMechanism.standard_output_ports>` of a `RecurrentTransferMechanism`:

COMMENT:
The two elements of the **MAX_VS_NEXT** OutputPort contain, respectively, the index of the LCAMechanism element with
the greatest value, and the difference between its value and the next highest one. **MAX_VS_AVG** OutputPort contains
the index of the LCAMechanism element with the greatest value, and the difference between its  value and the average
of all the others.
COMMENT
The `value <OutputPort.value>` of the *MAX_VS_NEXT* OutputPort contains the difference between the two elements of
the LCAMechanism’s `value <Mechanism_Base.value>` with the highest values, and the `value <OutputPort.value>` of the
*MAX_VS_AVG* OutputPort contains the difference between the element with the highest value and the average of all
the others (see `above <LCAMechanism_DDM_APPROXIMATION>` for their relationship to the output of a `DDM` Mechanism).

.. _LCAMechanism_Execution:

Execution
---------

The execution of an LCAMechanism is identical to that of `RecurrentTransferMechanism`.

.. _LCAMechanism_Class_Reference:

Class Reference
---------------


"""

import logging
import warnings
import sys

from collections.abc import Iterable

import numpy as np
from beartype import beartype

from psyneulink._typing import Optional, Union

from psyneulink.core.components.functions.nonstateful.objectivefunctions import Distance, MAX_ABS_DIFF
from psyneulink.core.components.functions.nonstateful.selectionfunctions import max_vs_avg, max_vs_next
from psyneulink.core.components.functions.stateful.integratorfunctions import LeakyCompetingIntegrator
from psyneulink.core.components.functions.nonstateful.transferfunctions import Logistic
from psyneulink.core.components.mechanisms.mechanism import MechanismError
from psyneulink.core.components.mechanisms.processing.transfermechanism import _integrator_mode_setter
from psyneulink.core.globals.keywords import \
    (CONVERGENCE, FUNCTION, GREATER_THAN_OR_EQUAL, LCA_MECHANISM, LESS_THAN_OR_EQUAL,
     MATRIX, MAX_VS_NEXT, MAX_VS_AVG, NAME, RESULT, TERMINATION_THRESHOLD, TERMINATION_MEASURE,
     TERMINATION_COMPARISION_OP, VALUE, INVERSE_HOLLOW_MATRIX, AUTO)
from psyneulink.core.globals.parameters import FunctionParameter, Parameter, check_user_specified
from psyneulink.core.globals.preferences.basepreferenceset import ValidPrefSet
from psyneulink.library.components.mechanisms.processing.transfer.recurrenttransfermechanism import \
    RecurrentTransferMechanism, _recurrent_transfer_mechanism_matrix_getter, _recurrent_transfer_mechanism_matrix_setter

__all__ = ['LCAMechanism', 'LCAError', 'CONVERGENCE']


logger = logging.getLogger(__name__)


class LCAError(MechanismError):
    pass



# IMPLEMENTATION NOTE:  IMPLEMENTS OFFSET PARAM BUT IT IS NOT CURRENTLY BEING USED
class LCAMechanism(RecurrentTransferMechanism):
    """
    LCAMechanism(                    \
        leak=0.5,                    \
        competition=1.0,             \
        self_excitation=0.0,         \
        time_step_size=0.1,          \
        threshold = None             \
        threshold_criterion = VALUE)

    Subclass of `RecurrentTransferMechanism` that implements a Leaky Competitive Accumulator.
    See `RecurrentTransferMechanism <RecurrentTransferMechanism_Class_Reference>` for additional
    arguments and attributes.

    Arguments
    ---------

    leak : value : default 0.5
        specifies the `leak <LeakyCompetingIntegrator.leak>` for the `LeakyCompetingIntegrator` Function
        (see `leak <LCAMechanism.leak>` for additional details).

    competition : value : default 1.0
        specifies the magnitude of the off-diagonal terms in the LCAMechanism's `recurrent_projection
        <RecurrentTransferMechanism.recurrent_projection>` (see `competition <LCAMechanism.competition>` for
        additional details).

    self_excitation : value : default 0.0
        specifies the magnidute of the diagonal terms in the LCAMechanism's `recurrent_projection
        <RecurrentTransferMechanism.recurrent_projection>` (see `self_excitation <LCAMechanism.self_excitation>` for
        additional details).

    time_step_size : float : default 0.1
        assigned as the `time_step_size <LeakyCompetingIntegrator.time_step_size>` parameter of the
        `LeakyCompetingIntegrator` Function (see `time_step_size <LCAMechanism.time_step_size>` for additional details).

    threshold : float or None : default None
        specifes the value at which the Mechanism's `is_finished` attribute is set to True
        (see `LCAMechanism_Threshold` for additional details).

    threshold_criterion : *VALUE*, *MAX_VS_NEXT*, *MAX_VS_AVG*, or *CONVERGENCE*
        specifies the criterion that is used to evaluate whether the threshold has been reached. If *MAX_VS_NEXT* or
        *MAX_VS_AVG* is specified, then the length of the LCAMCechanism's `value <Mechanism_Base.value>` must be at
        least 2 (see `LCAMechanism_Threshold` for additional details).

    Attributes
    ----------

    matrix : 2d np.array
        the `matrix <MappingProjection.matrix>` parameter of the `recurrent_projection
        <RecurrentTransferMechanism.recurrent_projection>` for the Mechanism, the `self_excitation
        <LCAMechanism.self_excitation>` attribute sets the values on the diagonal, and the `competition
        <LCAMechanism.competition>` attribute sets the magnitude of the negative off-diagonal values.

    leak : value
        determines the `leak <LeakyCompetingIntegrator.leak>` for the `LeakyCompetingIntegrator` Function,
        which scales the contribution of its `previous_value <LeakyCompetingIntegrator.previous_value>` to the
        accumulation of its `variable <LeakyCompetingIntegrator.variable>` (:math:`x_{i}`) on each time step (see
        `LeakyCompetingIntegrator` for additional details.

    competition : value
        determines the magnitude of the off-diagonal terms in the LCAMechanism's `recurrent_projection
        <RecurrentTransferMechanism.recurrent_projection>`, thereby scaling the contributions of the competing unit
        (all :math:`f(x)_{j}` where :math:`j \\neq i`) to the accumulation of the `LeakyCompetingIntegrator's
        <LeakyCompetingIntegrator>` `variable <LeakyCompetingIntegrator.variable>` (:math:`x_{i}`) on each time step
        (see `LeakyCompetingIntegrator` for additional details.

    self_excitation : value
        determines the diagonal terms in the LCAMechanism's `recurrent_projection
        <RecurrentTransferMechanism.recurrent_projection>`, thereby scaling the contributions of each unit's own
        recurrent value (:math:`f(x)_{i}`) to the accumulation of the `LeakyCompetingIntegrator's
        <LeakyCompetingIntegrator>` `variable <LeakyCompetingIntegrator.value>` (:math:`x_{i}`) on each time step
        (see `LeakyCompetingIntegrator` for additional details.

    time_step_size : float
        parameter of the `LeakyCompetingIntegrator` Function that determines the timing precision of the integration
        process it implements, and used to scale its `noise <LeakyCompetingIntegrator.noise>` parameter appropriately.

    standard_output_ports : list[str]
        list of `Standard OutputPorts <OutputPort_Standard>` that includes the following in addition to the
        `standard_output_ports <RecurrentTransferMechanism.standard_output_ports>` of a `RecurrentTransferMechanism`:

        .. _MAX_VS_NEXT:

        *MAX_VS_NEXT* : float
            the difference between the two elements of the LCAMechanism's `value <Mechanism_Base.value>`
            with the highest values.

        .. _MAX_VS_AVG:

        *MAX_VS_AVG* : float
            the difference between the element of the LCAMechanism's `value <Mechanism_Base.value>`
            and the average of all of the other elements.


    Returns
    -------
    instance of LCAMechanism : LCAMechanism

    """
    componentType = LCA_MECHANISM

    class Parameters(RecurrentTransferMechanism.Parameters):
        """
            Attributes
            ----------

                auto
                    see `auto <LCAMechanism.auto>`

                    :default value: 0.0
                    :type: ``float``

                competition
                    see `competition <LCAMechanism.competition>`

                    :default value: 1.0
                    :type: ``float``

                function
                    see `function <LCAMechanism.function>`

                    :default value: `Logistic`
                    :type: `Function`

                hetero
                    see `hetero <LCAMechanism.hetero>`

                    :default value: -1.0
                    :type: ``float``

                initial_value
                    see `initial_value <LCAMechanism.initial_value>`

                    :default value: None
                    :type:

                integrator_function
                    see `integrator_function <LCAMechanism.integrator_function>`

                    :default value: `LeakyCompetingIntegrator`
                    :type: `Function`

                integrator_mode
                    see `integrator_mode <LCAMechanism_Integrator_Mode>`

                    :default value: True
                    :type: ``bool``

                leak
                    see `leak <LCAMechanism.leak>`

                    :default value: 0.5
                    :type: ``float``

                termination_measure
                    see `termination_measure <LCAMechanism.termination_measure>`

                    :default value: ``max``
                    :type: ``types.FunctionType``

                time_step_size
                    see `time_step_size <LCAMechanism.time_step_size>`

                    :default value: 0.1
                    :type: ``float``
        """

        function = Parameter(Logistic, stateful=False, loggable=False)

        matrix = Parameter(
            INVERSE_HOLLOW_MATRIX,
            modulable=False,
            getter=_recurrent_transfer_mechanism_matrix_getter,
            setter=_recurrent_transfer_mechanism_matrix_setter
        )
        integration_rate = FunctionParameter(
            0.5,
            function_name='integrator_function',
            function_parameter_name='rate',
            aliases='leak'
        )
        auto = Parameter(0.0, modulable=False, aliases='self_excitation')
        hetero = Parameter(-1.0, modulable=False)
        competition = Parameter(1.0, modulable=False)
        time_step_size = FunctionParameter(0.1, function_name='integrator_function')

        integrator_mode = Parameter(True, setter=_integrator_mode_setter, valid_types=bool)
        integrator_function = Parameter(LeakyCompetingIntegrator, stateful=False, loggable=False)
        termination_measure = Parameter(max, stateful=False, loggable=False)

        output_ports = Parameter(
            [RESULT],
            stateful=False,
            loggable=False,
            read_only=True,
            structural=True,
        )

        def _validate_competition(self, competition):
            if competition < 0:
                warnings.warn(
                    f"The 'competition' arg specified for {self._owner.name} is a negative value ({competition}); "
                    f"note that this will result in a matrix that has positive off-diagonal elements "
                    f"since 'competition' is assumed to specify the magnitude of inhibition."
                )

            return None

        def _validate_integration_rate(self, integration_rate):
            pass

    standard_output_ports = RecurrentTransferMechanism.standard_output_ports.copy()
    standard_output_ports.extend([{NAME:MAX_VS_NEXT,
                                    FUNCTION:max_vs_next},
                                   {NAME:MAX_VS_AVG,
                                    FUNCTION:max_vs_avg}])

    @check_user_specified
    @beartype
    def __init__(self,
                 default_variable=None,
                 input_shapes: Optional[Union[int, list, np.ndarray]] = None,
                 input_ports: Optional[Union[list, dict]] = None,
                 function=None,
                 initial_value=None,
                 leak=None,
                 competition=None,
                 hetero=None,
                 self_excitation=None,
                 noise=None,
                 integrator_mode=None,
                 time_step_size=None,
                 clip=None,
                 output_ports: Optional[Union[str, Iterable]] = None,
                 integrator_function=None,
                 params=None,
                 name=None,
                 prefs: Optional[ValidPrefSet] = None,
                 **kwargs):
        """Instantiate LCAMechanism
        """
        # MODIFIED 1/22/20 NEW: [JDC]
        if MATRIX in kwargs:
            matrix = kwargs[MATRIX]
            if matrix is not None:
                self_excitation = None
                competition = None
        else:
            matrix = None
        # MODIFIED 1/22/20 END

        try:
            if self_excitation is not None and kwargs[AUTO] is not None:
                if kwargs[AUTO] != self_excitation:
                    raise LCAError(
                        'Both self_excitation and auto are specified with'
                        ' different values. self_excitation is an alias of auto.'
                        ' Provided values: self_excitation = {0} , auto = {1}'.format(
                            self_excitation,
                            kwargs[AUTO]
                        )
                    )
            elif kwargs[AUTO] is not None:
                self_excitation = kwargs[AUTO]
                del kwargs[AUTO]
        except KeyError:
            pass

        if competition is not None and hetero is not None:
            if competition != -1.0 * hetero:
                raise LCAError(
                    'Both competition and hetero are specified. competition '
                    'must have the same magnitude but opposite sign of hetero. '
                    'Provided values: competition = {0} , hetero = {1}'.format(
                        competition,
                        hetero
                    )
                )
        elif competition is not None:
            hetero = -competition
        elif hetero is not None:
            competition = -hetero

        # MODIFIED 10/26/19 NEW: [JDC]
        # Implemented for backward compatibility or, if kept, ease of use
        termination_threshold, termination_measure, termination_comparison_op = self._parse_threshold_args(kwargs)
        # MODIFIED 10/26/19 END

        # Set maximum executions absurdly large to avoid early termination
        self.max_executions_before_finished = sys.maxsize

        super().__init__(
            default_variable=default_variable,
            input_shapes=input_shapes,
            input_ports=input_ports,
            # matrix=matrix,
            auto=self_excitation,
            hetero=hetero,
            function=function,
            integrator_function=integrator_function,
            initial_value=initial_value,
            noise=noise,
            clip=clip,
            termination_threshold=termination_threshold,
            termination_measure=termination_measure,
            termination_comparison_op=termination_comparison_op,
            leak=leak,
            self_excitation=self_excitation,
            competition=competition,
            integrator_mode=integrator_mode,
            time_step_size=time_step_size,
            output_ports=output_ports,
            params=params,
            name=name,
            prefs=prefs,
            **kwargs
        )

        # Do these here so that name of the object (assigned by super) can be used in the warning messages
        if matrix is not None:
            # # MODIFIED 1/22/20 OLD:
            # warnings.warn(f"The 'matrix' arg was specified for {self.name} but will not be used; "
            #               f"the matrix for an {self.__class__.__name__} is specified using "
            #               f"the 'self_excitation' and 'competition' args.")
            # MODIFIED 1/22/20 NEW: [JDC]
            warnings.warn(f"The 'matrix' arg was specified for {self.name}, "
                          f"so its 'self_excitation' and 'competition' arguments will be ignored.")
            # MODIFIED 1/22/20 END

    def _parse_threshold_args(self, kwargs):
        """Implements convenience arguments threshold and threshold_criterion

        These are translated into the appropriate specifications for the termination_threshold, termination_measure,
        and termination_comparison_op for TransferMechanism.

        Note:  specifying (threshold and termination_threshold) and/or (threshold and
        threshold_criterion and termination_measure) causes an error.
        """
        termination_threshold = kwargs.pop(TERMINATION_THRESHOLD, None)
        threshold = kwargs.pop('threshold', None)
        if threshold is not None:
            if termination_threshold is not None:
                raise LCAError(f"The {repr('threshold')} and {repr(TERMINATION_THRESHOLD)} "
                               f"args of {self.__class__} can't both be specified.")
            else:
                termination_threshold = threshold
        else:
            termination_threshold = termination_threshold or self.parameters.termination_threshold.default_value

        termination_measure = kwargs.pop(TERMINATION_MEASURE, None)
        termination_comparison_op = kwargs.pop(TERMINATION_COMPARISION_OP, None)
        threshold_criterion = kwargs.pop('threshold_criterion', None)
        if threshold_criterion is not None:
            if termination_measure is not None:
                raise LCAError(f"The {repr('threshold_criterion')} and {repr(TERMINATION_MEASURE)} "
                               f"args of {self.__name__} can't both be specified.")
            else:
                if threshold_criterion == VALUE:
                    termination_measure = max
                    termination_comparison_op = GREATER_THAN_OR_EQUAL
                elif threshold_criterion == MAX_VS_NEXT:
                    termination_measure = max_vs_next
                    termination_comparison_op = GREATER_THAN_OR_EQUAL
                elif threshold_criterion == MAX_VS_AVG:
                    termination_measure = max_vs_avg
                    termination_comparison_op = GREATER_THAN_OR_EQUAL
                elif threshold_criterion == CONVERGENCE:
                    termination_measure = Distance(metric=MAX_ABS_DIFF)
                    termination_comparison_op = LESS_THAN_OR_EQUAL
                else:
                    raise LCAError(f"Unrecognized value provided to 'threshold_criterion arg of {self.__name__}: "
                                   f"{threshold_criterion};  must be VALUE, MAX_VS_NEXT, MAX_VS_AVG or CONVERGENCE")
        else:
            termination_measure = termination_measure or max
            termination_comparison_op = termination_comparison_op or GREATER_THAN_OR_EQUAL

        return termination_threshold, termination_measure, termination_comparison_op



import functools

import numpy as np

import psyneulink as pnl
import psyneulink.core.components.functions.nonstateful.transferfunctions


def my_linear_fct(x,
                  m=2.0,
                  b=0.0,
                  params={pnl.ADDITIVE_PARAM:'b',
                          pnl.MULTIPLICATIVE_PARAM:'m'}):
    return m * x + b

def my_simple_linear_fct(x,
                         m=1.0,
                         b=0.0
                         ):
    return m * x + b

def my_exp_fct(x,
               r=1.0,
               # b=pnl.CONTROL,
               b=0.0,
               params={pnl.ADDITIVE_PARAM:'b',
                       pnl.MULTIPLICATIVE_PARAM:'r'}
               ):
    return x**r + b

def my_sinusoidal_fct(input,
                      phase=0,
                      amplitude=1,
                      params={pnl.ADDITIVE_PARAM:'phase',
                              pnl.MULTIPLICATIVE_PARAM:'amplitude'}):
    frequency = input[0]
    t = input[1]
    return amplitude * np.sin(2 * np.pi * frequency * t + phase)

Input_Layer = pnl.TransferMechanism(
    name='Input_Layer',
    default_variable=np.zeros((2,)),
    function=psyneulink.core.components.functions.nonstateful.transferfunctions.Logistic
)

Output_Layer = pnl.TransferMechanism(
        name='Output_Layer',
        default_variable=[0, 0, 0],
        function=psyneulink.core.components.functions.nonstateful.transferfunctions.Linear,
        # function=pnl.Logistic,
        # output_ports={pnl.NAME: 'RESULTS USING UDF',
        #                pnl.VARIABLE: [(pnl.OWNER_VALUE,0), pnl.TIME_STEP],
        #                pnl.FUNCTION: my_sinusoidal_fct}
        output_ports={pnl.NAME: 'RESULTS USING UDF',
                       # pnl.VARIABLE: (pnl.OWNER_VALUE, 0),
                       pnl.FUNCTION: psyneulink.core.components.functions.nonstateful.transferfunctions.Linear(slope=pnl.GATE)
                       # pnl.FUNCTION: pnl.Logistic(gain=pnl.GATING)
                       # pnl.FUNCTION: my_linear_fct
                       # pnl.FUNCTION: my_exp_fct
                       # pnl.FUNCTION:pnl.UserDefinedFunction(custom_function=my_simple_linear_fct,
                       #                                      params={pnl.ADDITIVE_PARAM:'b',
                       #                                              pnl.MULTIPLICATIVE_PARAM:'m',
                       #                                              },
                                                            # m=pnl.GATING,
                                                            # b=2.0
                                                            # )
                       }
)

Gating_Mechanism = pnl.GatingMechanism(
    # default_gating_allocation=0.0,
    input_shapes=[1],
    gating_signals=[
        # Output_Layer
        Output_Layer.output_port,
    ]
)


def print_header(system):
    print("\n\n**** Time: ", system.scheduler.get_clock(system).simple_time)


def show_target(context=None):
    print('Gated: ',
          Gating_Mechanism.gating_signals[0].efferents[0].receiver.owner.name,
          Gating_Mechanism.gating_signals[0].efferents[0].receiver.name)
    print('- Input_Layer.value:                  ', Input_Layer.parameters.value.get(context))
    print('- Output_Layer.value:                 ', Output_Layer.parameters.value.get(context))
    print('- Output_Layer.output_port.variable: ', Output_Layer.output_port.parameters.variable.get(context))
    print('- Output_Layer.output_port.value:    ', Output_Layer.output_port.parameters.value.get(context))

stim_list = {
    Input_Layer: [[-1, 30], [-1, 30], [-1, 30], [-1, 30]],
    Gating_Mechanism: [[0.0], [0.5], [1.0], [2.0]]
}
comp = pnl.Composition(pathways=[[Input_Layer, Output_Layer],[Gating_Mechanism]])
comp.show_graph()
comp.run(num_trials=4,
         inputs=stim_list,
         call_before_trial=functools.partial(print_header, comp),
         call_after_trial=show_target,
)



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# *********************************************  ControlProjection *********************************************************

"""

Contents
--------

  * `ControlProjection_Overview`
  * `ControlProjection_Creation`
      - `ControlProjection_Deferred_Initialization`
  * `ControlProjection_Structure`
  * `ControlProjection_Execution`
  * `ControlProjection_Class_Reference`


.. _ControlProjection_Overview:

Overview
--------

A ControlProjection is a type of `ModulatoryProjection <ModulatoryProjection>` that projects to the `ParameterPort
<ParameterPort>` of a `ProcessingMechanism <ProcessingMechanism>`. It takes the `value <ControlSignal.value>` of a
`ControlSignal` of a `ControlMechanism <ControlMechanism>` and uses it to  modify the value of the parameter associated
with the ParameterPort to which it projects.  All of the ControlProjections in a Composition, along with its other
`control components <ControlMechanism>`, can be displayed using the Composition's `show_graph
<ShowGraph_show_graph_Method>` method with its **show_control** argument assigned as `True`.

.. _ControlProjection_Creation:

Creating a ControlProjection
----------------------------

A ControlProjection can be created using any of the standard ways to `create a Projection <Projection_Creation>`,
or by including it in a `tuple <ParameterPort_Tuple_Specification>` that specifies a parameter for a `Mechanism
<Mechanism>`, `MappingProjection`, or the `function <Component.function>` of either of these.  If a ControlProjection
is created explicitly (using its constructor), and either its **receiver** or **sender** argument is not specified,
its initialization is `deferred <ControlProjection_Deferred_Initialization>`.  If it is included in a `parameter
specification <ParameterPort_Specification>`, the `ParameterPort` for the parameter being specified will be assigned
as the ControlProjection's `receiver <ControlProjection.receiver>`. If the **receiver** belongs to a Mechanism that
is part of a `Composition`, then the ControlProjection's `sender <ControlProjection.sender>` is assigned to a
`ControlSignal` of the Composition's `controller <Composition.controller>`.  Otherwise, its initialization is
`deferred <ControlProjection_Deferred_Initialization>`.

.. _ControlProjection_Deferred_Initialization:

*Deferred Initialization*
~~~~~~~~~~~~~~~~~~~~~~~~~

When a ControlProjection is created, its full initialization is `deferred <Component_Deferred_Init>` until its `sender
<ControlProjection.sender>` and `receiver <ControlProjection.receiver>` have been fully specified.  This allows
a ControlProjection to be created before its `sender <ControlProjection.sender>` and/or `receiver
<ControlProjection.receiver>` have been created (e.g., before them in a script), by calling its constructor without
specifying its **sender** or **receiver** arguments. However, for the ControlProjection to be operational,
initialization must be completed by a call to its `deferred_init` method. This is done automatically if the
ControlProjection is included in a `tuple specification <ParameterPort_Tuple_Specification>` for the parameter of a
`Mechanism <Mechanism>` or its `function <Mechanism_Base.function>`, when the `ControlMechanism <ControlMechanism>`
is created for the `Composition` to which the parameter's owner belongs (see `ControlMechanism_Creation`).


.. _ControlProjection_Structure:

Structure
---------

The `sender <ControlProjection.sender>` of a ControlProjection is a `ControlSignal` of a `ControlMechanism
<ControlMechanism>`. The `value <ControlSignal.value>` of the `sender <ControlProjection.sender>` is used by the
ControlProjection as its `variable <ControlProjection.variable>`;  this is also assigned to its `control_signal
<ControlProjection.control_signal>` attribute, and serves as the input to the ControlProjection's `function
<Projection_Base.function>`.  The default `function <Projection_Base.function>` for a ControlProjection is an identity
function (`Linear` with **slope**\\ =1 and **intercept**\\ =0);  that is, it simply conveys the value of its
`control_signal <ControlProjection.control_signal>` to its `receiver <ControlProjection.receiver>`, for use in
modifying the value of the parameter that it controls. Its `receiver <ControlProjection.receiver>` is the
`ParameterPort` for the parameter of the `Mechanism <Mechanism>` or its `function Mechanism_Base.function>` that is
controlled by the ControlProjection.

.. _ControlProjection_Execution:

Execution
---------

A ControlProjection cannot be executed directly.  It is executed when the `ParameterPort` to which it projects is
updated.  Note that this only occurs when the `Mechanism <Mechanism>` to which the `ParameterPort` belongs is executed
(see `Lazy Evaluation <Component_Lazy_Updating>` for an explanation of "lazy" updating). When a ControlProjection is
executed, its `function <Projection_Base.function>` gets the `control_signal <ControlProjection.control_signal>` from
its `sender <ControlProjection.sender>` and conveys that to its `receiver <ControlProjection.receiver>`.  This is used
by the `receiver <ControlProjection.receiver>` to modify the parameter controlled by the ControlProjection (see
`ModulatorySignal_Modulation` and `ParameterPort Execution <ParameterPort_Execution>` for how modulation operates and
how this applies to a ParameterPort).

.. note::
   The changes to a parameter in response to the execution of a ControlProjection are not applied until the `Mechanism
   <Mechanism>` that receives the ControlProjection are next executed; see `Lazy Evaluation <Component_Lazy_Updating>`
   for an explanation of "lazy" updating).

.. _ControlProjection_Class_Reference:

Class Reference
---------------

"""

import inspect

from beartype import beartype

from psyneulink._typing import Mapping, Optional

from psyneulink.core.components.component import parameter_keywords
from psyneulink.core.components.functions.nonstateful.transferfunctions import Linear
from psyneulink.core.components.mechanisms.modulatory.control.controlmechanism import ControlMechanism
from psyneulink.core.components.projections.modulatory.modulatoryprojection import ModulatoryProjection_Base
from psyneulink.core.components.projections.projection import ProjectionError, projection_keywords
from psyneulink.core.components.shellclasses import Mechanism, Process_Base
from psyneulink.core.globals.context import ContextFlags
from psyneulink.core.globals.keywords import \
    CONTROL, CONTROL_PROJECTION, CONTROL_SIGNAL, INPUT_PORT, OUTPUT_PORT, PARAMETER_PORT
from psyneulink.core.globals.parameters import Parameter, SharedParameter, check_user_specified
from psyneulink.core.globals.preferences.basepreferenceset import ValidPrefSet
from psyneulink.core.globals.preferences.preferenceset import PreferenceLevel

__all__ = [
    'CONTROL_SIGNAL_PARAMS', 'ControlProjection', 'ControlProjectionError',
]

parameter_keywords.update({CONTROL_PROJECTION, CONTROL})
projection_keywords.update({CONTROL_PROJECTION, CONTROL})

CONTROL_SIGNAL_PARAMS = 'control_signal_params'

class ControlProjectionError(ProjectionError):
    pass


class ControlProjection(ModulatoryProjection_Base):
    """
    ControlProjection(           \
     sender=None,                \
     receiver=None,              \
     control_signal_params=None)

    Subclass of `ModulatoryProjection <ModulatoryProjection>` that modulates the value of an `InputPort`,
    `ParameterPort`, or `OutputPort` of a `Mechanism <Mechanism>`.
    See `Projection <ModulatoryProjection_Class_Reference>` for additional arguments and attributes.

    Arguments
    ---------

    sender : ControlMechanism or ControlSignal : default None
        specifies the source of the `control_signal <ControlProjection.control_signal>` for the ControlProjection;
        if it is not specified and cannot be `inferred from context <ControlProjection_Creation>`, initialization is
        `deferred <ControlProjection_Deferred_Initialization>`.

    receiver : Mechanism or ParameterPort  : default None
        specifies the `InputPort`, `ParameterPort` or `OutputPort` associated with the parameter to be controlled; if
        it is not specified, and cannot be `inferred from context <ControlProjection_Creation>`, initialization is
        `deferred <ControlProjection_Deferred_Initialization>`.

    control_signal_params : Dict[param keyword: param value] : None
        a `parameter dictionary <ParameterPort_Specification>` that can be used to specify the parameters for the
        ControlProjection's `sender <ControlProjection.sender>` (see `ControlSignal_Structure` for a description
        of ControlSignal parameters).

    Attributes
    ----------

    sender : ControlSignal
        source of the `control_signal <ControlProjection.control_signal>`.

    receiver : ParameterPort of Mechanism
        `ParameterPort` for the parameter to be modified by the ControlProjection.

    variable : 2d np.array
        same as `control_signal <ControlProjection.control_signal>`.

    control_signal : 1d np.array
        the `value <ControlSignal.value>` of the ControlProjection's `sender <ControlProjection.sender>`.

    value : float
        the value used to modify the parameter controlled by the ControlProjection (see `ModulatorySignal_Modulation`
        and `ParameterPort Execution <ParameterPort_Execution>` for how modulation operates and how this applies
        to a ParameterPort).

    """

    color = 0

    componentType = CONTROL_PROJECTION
    className = componentType
    suffix = " " + className

    classPreferenceLevel = PreferenceLevel.TYPE

    class sockets:
        sender=[CONTROL_SIGNAL]
        receiver=[PARAMETER_PORT, INPUT_PORT, OUTPUT_PORT]

    class Parameters(ModulatoryProjection_Base.Parameters):
        """
            Attributes
            ----------

                control_signal
                    see `control_signal <ControlProjection.control_signal>`

                    :default value: None
                    :type:
                    :read only: True

                function
                    see `function <ControlProjection.function>`

                    :default value: `Linear`
                    :type: `Function`
        """
        function = Parameter(Linear, stateful=False, loggable=False)
        control_signal = SharedParameter(None, attribute_name='sender', shared_parameter_name='value')

        control_signal_params = Parameter(
            None,
            stateful=False,
            loggable=False,
            read_only=True,
            user=False,
            pnl_internal=True
        )


    projection_sender = ControlMechanism

    @check_user_specified
    @beartype
    def __init__(self,
                 sender=None,
                 receiver=None,
                 weight=None,
                 exponent=None,
                 function=None,
                 control_signal_params:Optional[Mapping]=None,
                 params=None,
                 name=None,
                 prefs:   Optional[ValidPrefSet] = None,
                 **kwargs):
        # If receiver has not been assigned, defer init to Port.instantiate_projection_to_state()
        if (sender is None or sender.initialization_status == ContextFlags.DEFERRED_INIT or
                inspect.isclass(receiver) or receiver is None or
                    receiver.initialization_status == ContextFlags.DEFERRED_INIT):
            self.initialization_status = ContextFlags.DEFERRED_INIT

        # Validate sender (as variable) and params, and assign to variable
        # Note: pass name of mechanism (to override assignment of componentName in super.__init__)
        # super(ControlSignal_Base, self).__init__(sender=sender,
        super(ControlProjection, self).__init__(
            sender=sender,
            receiver=receiver,
            weight=weight,
            exponent=exponent,
            function=function,
            control_signal_params=control_signal_params,
            params=params,
            name=name,
            prefs=prefs,
            **kwargs
        )

    def _instantiate_sender(self, sender, params=None, context=None):
        """Check if DefaultController is being assigned and if so configure it for the requested ControlProjection

        If self.sender is a Mechanism, re-assign to <Mechanism>.outputPort
        Insure that sender.value = self.defaults.variable

        This method overrides the corresponding method of Projection, before calling it, to check if the
            DefaultController is being assigned as sender and, if so:
            - creates projection-dedicated inputPort and outputPort in DefaultController
            - puts them in DefaultController's input_ports and outputPorts attributes
            - lengthens variable of DefaultController to accommodate the ControlProjection
            - updates value of the DefaultController (in response to the new variable)
        Notes:
            * the default function of the DefaultControlMechanism simply maps the inputPort value to the outputPort
            * the params arg is assumed to be a dictionary of params for the ControlSignal of the ControlMechanism

        :return:
        """

        # A Process can't be the sender of a ControlMechanism
        if isinstance(sender, Process_Base):
            raise ProjectionError(
                "PROGRAM ERROR: attempt to add a ControlProjection from a Process {0} "
                "to a Mechanism {0} in pathway list".format(self.name, sender.name)
            )

        # If sender is specified as a Mechanism, validate that it is a ControlMechanism
        if isinstance(sender, Mechanism):
            # If sender is a ControlMechanism, call it to instantiate its ControlSignal projection
            if not isinstance(sender, ControlMechanism):
                raise ControlProjectionError(
                    "Mechanism specified as sender for {} ({}) must be a "
                    "ControlMechanism (but it is a {})".format(
                        self.name, sender.name, sender.__class__.__name__
                    )
                )

        # Call super to instantiate sender
        super()._instantiate_sender(sender, context=context)


    def _instantiate_receiver(self, context=None):
        # FIX: THIS NEEDS TO BE PUT BEFORE _instantiate_function SINCE THAT USES self.receiver
        """Handle situation in which self.receiver was specified as a Mechanism (rather than Port)

        Overrides Projection._instantiate_receiver, to require that if the receiver is specified as a Mechanism, then:
            the receiver Mechanism must have one and only one ParameterPort;
            otherwise, passes control to Projection_Base._instantiate_receiver for validation

        :return:
        """
        if isinstance(self.receiver, Mechanism):
            # If there is just one param of ParameterPort type in the receiver Mechanism
            # then assign it as actual receiver (which must be a Port);  otherwise, raise exception
            if len(self.receiver.parameter_ports) == 1:
                # Reassign self.receiver to Mechanism's parameterPort
                self.receiver = self.receiver.parameter_ports[0]
            else:
                raise ControlProjectionError("Unable to assign ControlProjection ({0}) from {1} to {2}, "
                                         "as it does not have exactly one ParameterPort;  must specify one (or each) of them"
                                         " as receiver(s)".
                                         format(self.name, self.sender.owner, self.receiver.name))
        # else:
        super(ControlProjection, self)._instantiate_receiver(context=context)

    # def _execute(self, variable=None, context=None, runtime_params=None):
    #     return super()._execute(variable, context, runtime_params)

    @property
    def control_signal(self):
        return self.sender.value
